from __future__ import annotations

import uuid
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Final,
    MutableMapping,
    TypedDict,
    Union,
    overload,
)
from weakref import WeakValueDictionary

import narwhals.stable.v1 as nw

from altair.utils._importers import import_vegafusion
from altair.utils.core import DataFrameLike
from altair.utils.data import (
    DataType,
    MaxRowsError,
    SupportsGeoInterface,
    ToValuesReturnType,
)
from altair.vegalite.data import default_data_transformer

if TYPE_CHECKING:
    from narwhals.typing import IntoDataFrame

    from vegafusion.runtime import ChartState  # type: ignore

# Temporary storage for dataframes that have been extracted
# from charts by the vegafusion data transformer. Use a WeakValueDictionary
# rather than a dict so that the Python interpreter is free to garbage
# collect the stored DataFrames.
extracted_inline_tables: MutableMapping[str, DataFrameLike] = WeakValueDictionary()

# Special URL prefix that VegaFusion uses to denote that a
# dataset in a Vega spec corresponds to an entry in the `inline_datasets`
# kwarg of vf.runtime.pre_transform_spec().
VEGAFUSION_PREFIX: Final = "vegafusion+dataset://"


class _ToVegaFusionReturnUrlDict(TypedDict):
    url: str


_VegaFusionReturnType = Union[_ToVegaFusionReturnUrlDict, ToValuesReturnType]


@overload
def vegafusion_data_transformer(
    data: None = ..., max_rows: int = ...
) -> Callable[..., Any]: ...


@overload
def vegafusion_data_transformer(
    data: DataFrameLike, max_rows: int = ...
) -> ToValuesReturnType: ...


@overload
def vegafusion_data_transformer(
    data: dict | IntoDataFrame | SupportsGeoInterface, max_rows: int = ...
) -> _VegaFusionReturnType: ...


def vegafusion_data_transformer(
    data: DataType | None = None, max_rows: int = 100000
) -> Callable[..., Any] | _VegaFusionReturnType:
    """VegaFusion Data Transformer."""
    # Vegafusion does not support Narwhals, so if `data` is a Narwhals
    # object, we make sure to extract the native object and let Vegafusion handle it.
    # `strict=False` passes `data` through as-is if it is not a Narwhals object.
    data = nw.to_native(data, strict=False)
    if data is None:
        return vegafusion_data_transformer
    elif isinstance(data, DataFrameLike) and not isinstance(data, SupportsGeoInterface):
        table_name = f"table_{uuid.uuid4()}".replace("-", "_")
        extracted_inline_tables[table_name] = data
        return {"url": VEGAFUSION_PREFIX + table_name}
    else:
        # Use default transformer for geo interface objects
        # # (e.g. a geopandas GeoDataFrame)
        # Or if we don't recognize data type
        return default_data_transformer(data)


def get_inline_table_names(vega_spec: dict[str, Any]) -> set[str]:
    """
    Get a set of the inline datasets names in the provided Vega spec.

    Inline datasets are encoded as URLs that start with the table://
    prefix.

    Parameters
    ----------
    vega_spec: dict
        A Vega specification dict

    Returns
    -------
    set of str
        Set of the names of the inline datasets that are referenced
        in the specification.

    Examples
    --------
    >>> spec = {
    ...     "data": [
    ...         {"name": "foo", "url": "https://path/to/file.csv"},
    ...         {"name": "bar", "url": "vegafusion+dataset://inline_dataset_123"},
    ...     ]
    ... }
    >>> get_inline_table_names(spec)
    {'inline_dataset_123'}
    """
    table_names = set()

    # Process datasets
    for data in vega_spec.get("data", []):
        url = data.get("url", "")
        if url.startswith(VEGAFUSION_PREFIX):
            name = url[len(VEGAFUSION_PREFIX) :]
            table_names.add(name)

    # Recursively process child marks, which may have their own datasets
    for mark in vega_spec.get("marks", []):
        table_names.update(get_inline_table_names(mark))

    return table_names


def get_inline_tables(vega_spec: dict[str, Any]) -> dict[str, DataFrameLike]:
    """
    Get the inline tables referenced by a Vega specification.

    Note: This function should only be called on a Vega spec that corresponds
    to a chart that was processed by the vegafusion_data_transformer.
    Furthermore, this function may only be called once per spec because
    the returned dataframes are deleted from internal storage.

    Parameters
    ----------
    vega_spec: dict
        A Vega specification dict

    Returns
    -------
    dict from str to dataframe
        dict from inline dataset name to dataframe object
    """
    inline_names = get_inline_table_names(vega_spec)
    # exclude named dataset that was provided by the user,
    # or dataframes that have been deleted.
    table_names = inline_names.intersection(extracted_inline_tables)
    return {k: extracted_inline_tables.pop(k) for k in table_names}


def compile_to_vegafusion_chart_state(
    vegalite_spec: dict[str, Any], local_tz: str
) -> ChartState:
    """
    Compile a Vega-Lite spec to a VegaFusion ChartState.

    Note: This function should only be called on a Vega-Lite spec
    that was generated with the "vegafusion" data transformer enabled.
    In particular, this spec may contain references to extract datasets
    using table:// prefixed URLs.

    Parameters
    ----------
    vegalite_spec: dict
        A Vega-Lite spec that was generated from an Altair chart with
        the "vegafusion" data transformer enabled
    local_tz: str
        Local timezone name (e.g. 'America/New_York')

    Returns
    -------
    ChartState
        A VegaFusion ChartState object
    """
    # Local import to avoid circular ImportError
    from altair import data_transformers, vegalite_compilers

    vf = import_vegafusion()

    # Compile Vega-Lite spec to Vega
    compiler = vegalite_compilers.get()
    if compiler is None:
        msg = "No active vega-lite compiler plugin found"
        raise ValueError(msg)

    vega_spec = compiler(vegalite_spec)

    # Retrieve dict of inline tables referenced by the spec
    inline_tables = get_inline_tables(vega_spec)

    # Pre-evaluate transforms in vega spec with vegafusion
    row_limit = data_transformers.options.get("max_rows", None)

    chart_state = vf.runtime.new_chart_state(
        vega_spec,
        local_tz=local_tz,
        inline_datasets=inline_tables,
        row_limit=row_limit,
    )

    # Check from row limit warning and convert to MaxRowsError
    handle_row_limit_exceeded(row_limit, chart_state.get_warnings())

    return chart_state


def compile_with_vegafusion(vegalite_spec: dict[str, Any]) -> dict[str, Any]:
    """
    Compile a Vega-Lite spec to Vega and pre-transform with VegaFusion.

    Note: This function should only be called on a Vega-Lite spec
    that was generated with the "vegafusion" data transformer enabled.
    In particular, this spec may contain references to extract datasets
    using table:// prefixed URLs.

    Parameters
    ----------
    vegalite_spec: dict
        A Vega-Lite spec that was generated from an Altair chart with
        the "vegafusion" data transformer enabled

    Returns
    -------
    dict
        A Vega spec that has been pre-transformed by VegaFusion
    """
    # Local import to avoid circular ImportError
    from altair import data_transformers, vegalite_compilers

    vf = import_vegafusion()

    # Compile Vega-Lite spec to Vega
    compiler = vegalite_compilers.get()
    if compiler is None:
        msg = "No active vega-lite compiler plugin found"
        raise ValueError(msg)

    vega_spec = compiler(vegalite_spec)

    # Retrieve dict of inline tables referenced by the spec
    inline_tables = get_inline_tables(vega_spec)

    # Pre-evaluate transforms in vega spec with vegafusion
    row_limit = data_transformers.options.get("max_rows", None)
    transformed_vega_spec, warnings = vf.runtime.pre_transform_spec(
        vega_spec,
        vf.get_local_tz(),
        inline_datasets=inline_tables,
        row_limit=row_limit,
    )

    # Check from row limit warning and convert to MaxRowsError
    handle_row_limit_exceeded(row_limit, warnings)

    return transformed_vega_spec


def handle_row_limit_exceeded(row_limit: int, warnings: list):
    for warning in warnings:
        if warning.get("type") == "RowLimitExceeded":
            msg = (
                "The number of dataset rows after filtering and aggregation exceeds\n"
                f"the current limit of {row_limit}. Try adding an aggregation to reduce\n"
                "the size of the dataset that must be loaded into the browser. Or, disable\n"
                "the limit by calling alt.data_transformers.disable_max_rows(). Note that\n"
                "disabling this limit may cause the browser to freeze or crash."
            )
            raise MaxRowsError(msg)


def using_vegafusion() -> bool:
    """Check whether the vegafusion data transformer is enabled."""
    # Local import to avoid circular ImportError
    from altair import data_transformers

    return data_transformers.active == "vegafusion"



from .core import (
    SHORTHAND_KEYS,
    SchemaBase,
    display_traceback,
    infer_encoding_types,
    infer_vegalite_type_for_pandas,
    parse_shorthand,
    sanitize_narwhals_dataframe,
    sanitize_pandas_dataframe,
    update_nested,
    use_signature,
)
from .deprecation import AltairDeprecationWarning, deprecated, deprecated_warn
from .html import spec_to_html
from .plugin_registry import PluginRegistry
from .schemapi import Optional, Undefined, is_undefined

__all__ = (
    "SHORTHAND_KEYS",
    "AltairDeprecationWarning",
    "Optional",
    "PluginRegistry",
    "SchemaBase",
    "Undefined",
    "deprecated",
    "deprecated_warn",
    "display_traceback",
    "infer_encoding_types",
    "infer_vegalite_type_for_pandas",
    "is_undefined",
    "parse_shorthand",
    "sanitize_narwhals_dataframe",
    "sanitize_pandas_dataframe",
    "spec_to_html",
    "update_nested",
    "use_signature",
)



from __future__ import annotations

from importlib.metadata import version as importlib_version
from typing import TYPE_CHECKING

from packaging.version import Version

if TYPE_CHECKING:
    from types import ModuleType


def import_vegafusion() -> ModuleType:
    min_version = "1.5.0"
    try:
        version = importlib_version("vegafusion")
        embed_version = importlib_version("vegafusion-python-embed")
        if version != embed_version or Version(version) < Version(min_version):
            msg = (
                "The versions of the vegafusion and vegafusion-python-embed packages must match\n"
                f"and must be version {min_version} or greater.\n"
                f"Found:\n"
                f" - vegafusion=={version}\n"
                f" - vegafusion-python-embed=={embed_version}\n"
            )
            raise RuntimeError(msg)
        import vegafusion as vf  # type: ignore

        return vf
    except ImportError as err:
        msg = (
            'The "vegafusion" data transformer and chart.transformed_data feature requires\n'
            f"version {min_version} or greater of the 'vegafusion-python-embed' and 'vegafusion' packages.\n"
            "These can be installed with pip using:\n"
            f'    pip install "vegafusion[embed]>={min_version}"\n'
            "Or with conda using:\n"
            f'    conda install -c conda-forge "vegafusion-python-embed>={min_version}" '
            f'"vegafusion>={min_version}"\n\n'
            f"ImportError: {err.args[0]}"
        )
        raise ImportError(msg) from err


def import_vl_convert() -> ModuleType:
    min_version = "1.6.0"
    try:
        version = importlib_version("vl-convert-python")
        if Version(version) < Version(min_version):
            msg = (
                f"The vl-convert-python package must be version {min_version} or greater. "
                f"Found version {version}"
            )
            raise RuntimeError(msg)
        import vl_convert as vlc

        return vlc
    except ImportError as err:
        msg = (
            f"The vl-convert Vega-Lite compiler and file export feature requires\n"
            f"version {min_version} or greater of the 'vl-convert-python' package. \n"
            f"This can be installed with pip using:\n"
            f'   pip install "vl-convert-python>={min_version}"\n'
            "or conda:\n"
            f'   conda install -c conda-forge "vl-convert-python>={min_version}"\n\n'
            f"ImportError: {err.args[0]}"
        )
        raise ImportError(msg) from err


def vl_version_for_vl_convert() -> str:
    from altair.vegalite import SCHEMA_VERSION

    # Compute VlConvert's vl_version string (of the form 'v5_2')
    # from SCHEMA_VERSION (of the form 'v5.2.0')
    return "_".join(SCHEMA_VERSION.split(".")[:2])


def import_pyarrow_interchange() -> ModuleType:
    min_version = "11.0.0"
    try:
        version = importlib_version("pyarrow")

        if Version(version) < Version(min_version):
            msg = (
                f"The pyarrow package must be version {min_version} or greater. "
                f"Found version {version}"
            )
            raise RuntimeError(msg)
        import pyarrow.interchange as pi

        return pi
    except ImportError as err:
        msg = (
            f"Usage of the DataFrame Interchange Protocol requires\n"
            f"version {min_version} or greater of the pyarrow package. \n"
            f"This can be installed with pip using:\n"
            f'   pip install "pyarrow>={min_version}"\n'
            "or conda:\n"
            f'   conda install -c conda-forge "pyarrow>={min_version}"\n\n'
            f"ImportError: {err.args[0]}"
        )
        raise ImportError(msg) from err


def pyarrow_available() -> bool:
    try:
        import_pyarrow_interchange()
        return True
    except (ImportError, RuntimeError):
        return False



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, overload

from altair.utils.core import sanitize_pandas_dataframe
from altair.utils.data import DataTransformerRegistry as _DataTransformerRegistry
from altair.utils.data import (
    MaxRowsError,
    check_data_type,
    limit_rows,
    sample,
    to_csv,
    to_json,
    to_values,
)

if TYPE_CHECKING:
    from altair.utils.data import DataType, ToValuesReturnType
    from altair.utils.plugin_registry import PluginEnabler


@overload
def default_data_transformer(
    data: None = ..., max_rows: int = ...
) -> Callable[[DataType], ToValuesReturnType]: ...
@overload
def default_data_transformer(
    data: DataType, max_rows: int = ...
) -> ToValuesReturnType: ...
def default_data_transformer(
    data: DataType | None = None, max_rows: int = 5000
) -> Callable[[DataType], ToValuesReturnType] | ToValuesReturnType:
    if data is None:

        def pipe(data: DataType, /) -> ToValuesReturnType:
            data = limit_rows(data, max_rows=max_rows)
            return to_values(data)

        return pipe

    else:
        return to_values(limit_rows(data, max_rows=max_rows))


class DataTransformerRegistry(_DataTransformerRegistry):
    def disable_max_rows(self) -> PluginEnabler:
        """Disable the MaxRowsError."""
        options = self.options
        if self.active in {"default", "vegafusion"}:
            options = options.copy()
            options["max_rows"] = None
        return self.enable(**options)


__all__ = (
    "DataTransformerRegistry",
    "MaxRowsError",
    "check_data_type",
    "default_data_transformer",
    "limit_rows",
    "sample",
    "sanitize_pandas_dataframe",
    "to_csv",
    "to_json",
    "to_values",
)



"""Tools for creating transform & filter expressions with a python syntax."""

from __future__ import annotations

import sys

from altair.expr.core import ConstExpression, FunctionExpression
from altair.vegalite.v5.schema.core import ExprRef as _ExprRef

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override


class _ConstExpressionType(type):
    """Metaclass providing read-only class properties for :class:`expr`."""

    @property
    def NaN(cls) -> ConstExpression:
        """Not a number (same as JavaScript literal NaN)."""
        return ConstExpression("NaN")

    @property
    def LN10(cls) -> ConstExpression:
        """The natural log of 10 (alias to Math.LN10)."""
        return ConstExpression("LN10")

    @property
    def E(cls) -> ConstExpression:
        """The transcendental number e (alias to Math.E)."""
        return ConstExpression("E")

    @property
    def LOG10E(cls) -> ConstExpression:
        """The base 10 logarithm e (alias to Math.LOG10E)."""
        return ConstExpression("LOG10E")

    @property
    def LOG2E(cls) -> ConstExpression:
        """The base 2 logarithm of e (alias to Math.LOG2E)."""
        return ConstExpression("LOG2E")

    @property
    def SQRT1_2(cls) -> ConstExpression:
        """The square root of 0.5 (alias to Math.SQRT1_2)."""
        return ConstExpression("SQRT1_2")

    @property
    def LN2(cls) -> ConstExpression:
        """The natural log of 2 (alias to Math.LN2)."""
        return ConstExpression("LN2")

    @property
    def SQRT2(cls) -> ConstExpression:
        """The square root of 2 (alias to Math.SQRT1_2)."""
        return ConstExpression("SQRT2")

    @property
    def PI(cls) -> ConstExpression:
        """The transcendental number pi (alias to Math.PI)."""
        return ConstExpression("PI")


class expr(_ExprRef, metaclass=_ConstExpressionType):
    r"""
    Utility providing *constants* and *classmethods* to construct expressions.

    `Expressions`_ can be used to write basic formulas that enable custom interactions.

    Alternatively, an `inline expression`_ may be defined via :class:`expr()`.

    Parameters
    ----------
    expr: str
        A `vega expression`_ string.

    Returns
    -------
    ``ExprRef``

    .. _Expressions:
        https://altair-viz.github.io/user_guide/interactions.html#expressions
    .. _inline expression:
       https://altair-viz.github.io/user_guide/interactions.html#inline-expressions
    .. _vega expression:
       https://vega.github.io/vega/docs/expressions/

    Examples
    --------
    >>> import altair as alt

    >>> bind_range = alt.binding_range(min=100, max=300, name="Slider value:  ")
    >>> param_width = alt.param(bind=bind_range, name="param_width")
    >>> param_color = alt.param(
    ...     expr=alt.expr.if_(param_width < 200, "red", "black"),
    ...     name="param_color",
    ... )
    >>> y = alt.Y("yval").axis(titleColor=param_color)

    >>> y
    Y({
      axis: {'titleColor': Parameter('param_color', VariableParameter({
        expr: if((param_width < 200),'red','black'),
        name: 'param_color'
      }))},
      shorthand: 'yval'
    })
    """

    @override
    def __new__(cls: type[_ExprRef], expr: str) -> _ExprRef:  # type: ignore[misc]
        # NOTE: `mypy<=1.10.1` is not consistent with typing spec
        # https://github.com/python/mypy/issues/1020
        # https://docs.python.org/3/reference/datamodel.html#object.__new__
        # https://typing.readthedocs.io/en/latest/spec/constructors.html#new-method
        return _ExprRef(expr=expr)

    @classmethod
    def if_(cls, *args) -> FunctionExpression:
        """
        If *test* is truthy, returns *thenValue*. Otherwise, returns *elseValue*.

        The *if* function is equivalent to the ternary operator `a ? b : c`.
        """
        return FunctionExpression("if", args)

    @classmethod
    def isArray(cls, *args) -> FunctionExpression:
        """Returns true if *value* is an array, false otherwise."""
        return FunctionExpression("isArray", args)

    @classmethod
    def isBoolean(cls, *args) -> FunctionExpression:
        """Returns true if *value* is a boolean (`true` or `false`), false otherwise."""
        return FunctionExpression("isBoolean", args)

    @classmethod
    def isDate(cls, *args) -> FunctionExpression:
        """
        Returns true if *value* is a Date object, false otherwise.

        This method will return false for timestamp numbers or date-formatted strings; it recognizes Date objects only.
        """
        return FunctionExpression("isDate", args)

    @classmethod
    def isDefined(cls, *args) -> FunctionExpression:
        """
        Returns true if *value* is a defined value, false if *value* equals `undefined`.

        This method will return true for `null` and `NaN` values.
        """
        return FunctionExpression("isDefined", args)

    @classmethod
    def isNumber(cls, *args) -> FunctionExpression:
        """
        Returns true if *value* is a number, false otherwise.

        `NaN` and `Infinity` are considered numbers.
        """
        return FunctionExpression("isNumber", args)

    @classmethod
    def isObject(cls, *args) -> FunctionExpression:
        """Returns true if *value* is an object (including arrays and Dates), false otherwise."""
        return FunctionExpression("isObject", args)

    @classmethod
    def isRegExp(cls, *args) -> FunctionExpression:
        """Returns true if *value* is a RegExp (regular expression) object, false otherwise."""
        return FunctionExpression("isRegExp", args)

    @classmethod
    def isString(cls, *args) -> FunctionExpression:
        """Returns true if *value* is a string, false otherwise."""
        return FunctionExpression("isString", args)

    @classmethod
    def isValid(cls, *args) -> FunctionExpression:
        """Returns true if *value* is not `null`, `undefined`, or `NaN`, false otherwise."""
        return FunctionExpression("isValid", args)

    @classmethod
    def toBoolean(cls, *args) -> FunctionExpression:
        """
        Coerces the input *value* to a string.

        Null values and empty strings are mapped to `null`.
        """
        return FunctionExpression("toBoolean", args)

    @classmethod
    def toDate(cls, *args) -> FunctionExpression:
        """
        Coerces the input *value* to a Date instance.

        Null values and empty strings are mapped to `null`.
        If an optional *parser* function is provided, it is used to perform date parsing, otherwise `Date.parse` is used.
        Be aware that `Date.parse` has different implementations across browsers!
        """
        return FunctionExpression("toDate", args)

    @classmethod
    def toNumber(cls, *args) -> FunctionExpression:
        """
        Coerces the input *value* to a number.

        Null values and empty strings are mapped to `null`.
        """
        return FunctionExpression("toNumber", args)

    @classmethod
    def toString(cls, *args) -> FunctionExpression:
        """
        Coerces the input *value* to a string.

        Null values and empty strings are mapped to `null`.
        """
        return FunctionExpression("toString", args)

    @classmethod
    def isNaN(cls, *args) -> FunctionExpression:
        """
        Returns true if *value* is not a number.

        Same as JavaScript's `isNaN`.
        """
        return FunctionExpression("isNaN", args)

    @classmethod
    def isFinite(cls, *args) -> FunctionExpression:
        """
        Returns true if *value* is a finite number.

        Same as JavaScript's `isFinite`.
        """
        return FunctionExpression("isFinite", args)

    @classmethod
    def abs(cls, *args) -> FunctionExpression:
        """
        Returns the absolute value of *value*.

        Same as JavaScript's `Math.abs`.
        """
        return FunctionExpression("abs", args)

    @classmethod
    def acos(cls, *args) -> FunctionExpression:
        """
        Trigonometric arccosine.

        Same as JavaScript's `Math.acos`.
        """
        return FunctionExpression("acos", args)

    @classmethod
    def asin(cls, *args) -> FunctionExpression:
        """
        Trigonometric arcsine.

        Same as JavaScript's `Math.asin`.
        """
        return FunctionExpression("asin", args)

    @classmethod
    def atan(cls, *args) -> FunctionExpression:
        """
        Trigonometric arctangent.

        Same as JavaScript's `Math.atan`.
        """
        return FunctionExpression("atan", args)

    @classmethod
    def atan2(cls, *args) -> FunctionExpression:
        """
        Returns the arctangent of *dy / dx*.

        Same as JavaScript's `Math.atan2`.
        """
        return FunctionExpression("atan2", args)

    @classmethod
    def ceil(cls, *args) -> FunctionExpression:
        """
        Rounds *value* to the nearest integer of equal or greater value.

        Same as JavaScript's `Math.ceil`.
        """
        return FunctionExpression("ceil", args)

    @classmethod
    def clamp(cls, *args) -> FunctionExpression:
        """Restricts *value* to be between the specified *min* and *max*."""
        return FunctionExpression("clamp", args)

    @classmethod
    def cos(cls, *args) -> FunctionExpression:
        """
        Trigonometric cosine.

        Same as JavaScript's `Math.cos`.
        """
        return FunctionExpression("cos", args)

    @classmethod
    def exp(cls, *args) -> FunctionExpression:
        """
        Returns the value of *e* raised to the provided *exponent*.

        Same as JavaScript's `Math.exp`.
        """
        return FunctionExpression("exp", args)

    @classmethod
    def floor(cls, *args) -> FunctionExpression:
        """
        Rounds *value* to the nearest integer of equal or lower value.

        Same as JavaScript's `Math.floor`.
        """
        return FunctionExpression("floor", args)

    @classmethod
    def hypot(cls, *args) -> FunctionExpression:
        """
        Returns the square root of the sum of squares of its arguments.

        Same as JavaScript's `Math.hypot`.
        """
        return FunctionExpression("hypot", args)

    @classmethod
    def log(cls, *args) -> FunctionExpression:
        """
        Returns the natural logarithm of *value*.

        Same as JavaScript's `Math.log`.
        """
        return FunctionExpression("log", args)

    @classmethod
    def max(cls, *args) -> FunctionExpression:
        """
        Returns the maximum argument value.

        Same as JavaScript's `Math.max`.
        """
        return FunctionExpression("max", args)

    @classmethod
    def min(cls, *args) -> FunctionExpression:
        """
        Returns the minimum argument value.

        Same as JavaScript's `Math.min`.
        """
        return FunctionExpression("min", args)

    @classmethod
    def pow(cls, *args) -> FunctionExpression:
        """
        Returns *value* raised to the given *exponent*.

        Same as JavaScript's `Math.pow`.
        """
        return FunctionExpression("pow", args)

    @classmethod
    def random(cls, *args) -> FunctionExpression:
        """
        Returns a pseudo-random number in the range `[0, 1]`.

        Same as JavaScript's `Math.random`.
        """
        return FunctionExpression("random", args)

    @classmethod
    def round(cls, *args) -> FunctionExpression:
        """
        Rounds *value* to the nearest integer.

        Same as JavaScript's `Math.round`.
        """
        return FunctionExpression("round", args)

    @classmethod
    def sin(cls, *args) -> FunctionExpression:
        """
        Trigonometric sine.

        Same as JavaScript's `Math.sin`.
        """
        return FunctionExpression("sin", args)

    @classmethod
    def sqrt(cls, *args) -> FunctionExpression:
        """
        Square root function.

        Same as JavaScript's `Math.sqrt`.
        """
        return FunctionExpression("sqrt", args)

    @classmethod
    def tan(cls, *args) -> FunctionExpression:
        """
        Trigonometric tangent.

        Same as JavaScript's `Math.tan`.
        """
        return FunctionExpression("tan", args)

    @classmethod
    def sampleNormal(cls, *args) -> FunctionExpression:
        """
        Returns a sample from a univariate `normal (Gaussian) probability distribution <https://en.wikipedia.org/wiki/Normal_distribution>`__ with specified *mean* and standard deviation *stdev*.

        If unspecified, the mean defaults to `0` and the standard deviation defaults to `1`.
        """
        return FunctionExpression("sampleNormal", args)

    @classmethod
    def cumulativeNormal(cls, *args) -> FunctionExpression:
        """
        Returns the value of the `cumulative distribution function <https://en.wikipedia.org/wiki/Cumulative_distribution_function>`__ at the given input domain *value* for a normal distribution with specified *mean* and standard deviation *stdev*.

        If unspecified, the mean defaults to `0` and the standard deviation defaults to `1`.
        """
        return FunctionExpression("cumulativeNormal", args)

    @classmethod
    def densityNormal(cls, *args) -> FunctionExpression:
        """
        Returns the value of the `probability density function <https://en.wikipedia.org/wiki/Probability_density_function>`__ at the given input domain *value*, for a normal distribution with specified *mean* and standard deviation *stdev*.

        If unspecified, the mean defaults to `0` and the standard deviation defaults to `1`.
        """
        return FunctionExpression("densityNormal", args)

    @classmethod
    def quantileNormal(cls, *args) -> FunctionExpression:
        """
        Returns the quantile value (the inverse of the `cumulative distribution function <https://en.wikipedia.org/wiki/Cumulative_distribution_function)>`__ for the given input *probability*, for a normal distribution with specified *mean* and standard deviation *stdev*.

        If unspecified, the mean defaults to `0` and the standard deviation defaults to `1`.
        """
        return FunctionExpression("quantileNormal", args)

    @classmethod
    def sampleLogNormal(cls, *args) -> FunctionExpression:
        """
        Returns a sample from a univariate `log-normal probability distribution <https://en.wikipedia.org/wiki/Log-normal_distribution>`__ with specified log *mean* and log standard deviation *stdev*.

        If unspecified, the log mean defaults to `0` and the log standard deviation defaults to `1`.
        """
        return FunctionExpression("sampleLogNormal", args)

    @classmethod
    def cumulativeLogNormal(cls, *args) -> FunctionExpression:
        """
        Returns the value of the `cumulative distribution function <https://en.wikipedia.org/wiki/Cumulative_distribution_function>`__ at the given input domain *value* for a log-normal distribution with specified log *mean* and log standard deviation *stdev*.

        If unspecified, the log mean defaults to `0` and the log standard deviation defaults to `1`.
        """
        return FunctionExpression("cumulativeLogNormal", args)

    @classmethod
    def densityLogNormal(cls, *args) -> FunctionExpression:
        """
        Returns the value of the `probability density function <https://en.wikipedia.org/wiki/Probability_density_function>`__ at the given input domain *value*, for a log-normal distribution with specified log *mean* and log standard deviation *stdev*.

        If unspecified, the log mean defaults to `0` and the log standard deviation defaults to `1`.
        """
        return FunctionExpression("densityLogNormal", args)

    @classmethod
    def quantileLogNormal(cls, *args) -> FunctionExpression:
        """
        Returns the quantile value (the inverse of the `cumulative distribution function <https://en.wikipedia.org/wiki/Cumulative_distribution_function)>`__ for the given input *probability*, for a log-normal distribution with specified log *mean* and log standard deviation *stdev*.

        If unspecified, the log mean defaults to `0` and the log standard deviation defaults to `1`.
        """
        return FunctionExpression("quantileLogNormal", args)

    @classmethod
    def sampleUniform(cls, *args) -> FunctionExpression:
        """
        Returns a sample from a univariate `continuous uniform probability distribution <https://en.wikipedia.org/wiki/Uniform_distribution_(continuous)>`__ over the interval `[min, max]`.

        If unspecified, *min* defaults to `0` and *max* defaults to `1`.
        If only one argument is provided, it is interpreted as the *max* value.
        """
        return FunctionExpression("sampleUniform", args)

    @classmethod
    def cumulativeUniform(cls, *args) -> FunctionExpression:
        """
        Returns the value of the `cumulative distribution function <https://en.wikipedia.org/wiki/Cumulative_distribution_function>`__ at the given input domain *value* for a uniform distribution over the interval `[min, max]`.

        If unspecified, *min* defaults to `0` and *max* defaults to `1`.
        If only one argument is provided, it is interpreted as the *max* value.
        """
        return FunctionExpression("cumulativeUniform", args)

    @classmethod
    def densityUniform(cls, *args) -> FunctionExpression:
        """
        Returns the value of the `probability density function <https://en.wikipedia.org/wiki/Probability_density_function>`__ at the given input domain *value*,  for a uniform distribution over the interval `[min, max]`.

        If unspecified, *min* defaults to `0` and *max* defaults to `1`.
        If only one argument is provided, it is interpreted as the *max* value.
        """
        return FunctionExpression("densityUniform", args)

    @classmethod
    def quantileUniform(cls, *args) -> FunctionExpression:
        """
        Returns the quantile value (the inverse of the `cumulative distribution function <https://en.wikipedia.org/wiki/Cumulative_distribution_function)>`__ for the given input *probability*, for a uniform distribution over the interval `[min, max]`.

        If unspecified, *min* defaults to `0` and *max* defaults to `1`.
        If only one argument is provided, it is interpreted as the *max* value.
        """
        return FunctionExpression("quantileUniform", args)

    @classmethod
    def now(cls, *args) -> FunctionExpression:
        """Returns the timestamp for the current time."""
        return FunctionExpression("now", args)

    @classmethod
    def datetime(cls, *args) -> FunctionExpression:
        """
        Returns a new `Date` instance.

        The *month* is 0-based, such that `1` represents February.
        """
        return FunctionExpression("datetime", args)

    @classmethod
    def date(cls, *args) -> FunctionExpression:
        """Returns the day of the month for the given *datetime* value, in local time."""
        return FunctionExpression("date", args)

    @classmethod
    def day(cls, *args) -> FunctionExpression:
        """Returns the day of the week for the given *datetime* value, in local time."""
        return FunctionExpression("day", args)

    @classmethod
    def dayofyear(cls, *args) -> FunctionExpression:
        """Returns the one-based day of the year for the given *datetime* value, in local time."""
        return FunctionExpression("dayofyear", args)

    @classmethod
    def year(cls, *args) -> FunctionExpression:
        """Returns the year for the given *datetime* value, in local time."""
        return FunctionExpression("year", args)

    @classmethod
    def quarter(cls, *args) -> FunctionExpression:
        """Returns the quarter of the year (0-3) for the given *datetime* value, in local time."""
        return FunctionExpression("quarter", args)

    @classmethod
    def month(cls, *args) -> FunctionExpression:
        """Returns the (zero-based) month for the given *datetime* value, in local time."""
        return FunctionExpression("month", args)

    @classmethod
    def week(cls, *args) -> FunctionExpression:
        """
        Returns the week number of the year for the given *datetime*, in local time.

        This function assumes Sunday-based weeks.
        Days before the first Sunday of the year are considered to be in week 0,
        the first Sunday of the year is the start of week 1,
        the second Sunday week 2, etc.
        """
        return FunctionExpression("week", args)

    @classmethod
    def hours(cls, *args) -> FunctionExpression:
        """Returns the hours component for the given *datetime* value, in local time."""
        return FunctionExpression("hours", args)

    @classmethod
    def minutes(cls, *args) -> FunctionExpression:
        """Returns the minutes component for the given *datetime* value, in local time."""
        return FunctionExpression("minutes", args)

    @classmethod
    def seconds(cls, *args) -> FunctionExpression:
        """Returns the seconds component for the given *datetime* value, in local time."""
        return FunctionExpression("seconds", args)

    @classmethod
    def milliseconds(cls, *args) -> FunctionExpression:
        """Returns the milliseconds component for the given *datetime* value, in local time."""
        return FunctionExpression("milliseconds", args)

    @classmethod
    def time(cls, *args) -> FunctionExpression:
        """Returns the epoch-based timestamp for the given *datetime* value."""
        return FunctionExpression("time", args)

    @classmethod
    def timezoneoffset(cls, *args) -> FunctionExpression:
        """Returns the timezone offset from the local timezone to UTC for the given *datetime* value."""
        return FunctionExpression("timezoneoffset", args)

    @classmethod
    def timeOffset(cls, *args) -> FunctionExpression:
        """
        Returns a new `Date` instance that offsets the given *date* by the specified time `unit <https://vega.github.io/vega/docs/api/time/#time-units>`__ in the local timezone.

        The optional *step* argument indicates the number of time unit steps to offset by (default 1).
        """
        return FunctionExpression("timeOffset", args)

    @classmethod
    def timeSequence(cls, *args) -> FunctionExpression:
        """
        Returns an array of `Date` instances from *start* (inclusive) to *stop* (exclusive), with each entry separated by the given time `unit <https://vega.github.io/vega/docs/api/time/#time-units>`__ in the local timezone.

        The optional *step* argument indicates the number of time unit steps to take between each sequence entry (default 1).
        """
        return FunctionExpression("timeSequence", args)

    @classmethod
    def utc(cls, *args) -> FunctionExpression:
        """Returns a timestamp for the given UTC date. The *month* is 0-based, such that `1` represents February."""
        return FunctionExpression("utc", args)

    @classmethod
    def utcdate(cls, *args) -> FunctionExpression:
        """Returns the day of the month for the given *datetime* value, in UTC time."""
        return FunctionExpression("utcdate", args)

    @classmethod
    def utcday(cls, *args) -> FunctionExpression:
        """Returns the day of the week for the given *datetime* value, in UTC time."""
        return FunctionExpression("utcday", args)

    @classmethod
    def utcdayofyear(cls, *args) -> FunctionExpression:
        """Returns the one-based day of the year for the given *datetime* value, in UTC time."""
        return FunctionExpression("utcdayofyear", args)

    @classmethod
    def utcyear(cls, *args) -> FunctionExpression:
        """Returns the year for the given *datetime* value, in UTC time."""
        return FunctionExpression("utcyear", args)

    @classmethod
    def utcquarter(cls, *args) -> FunctionExpression:
        """Returns the quarter of the year (0-3) for the given *datetime* value, in UTC time."""
        return FunctionExpression("utcquarter", args)

    @classmethod
    def utcmonth(cls, *args) -> FunctionExpression:
        """Returns the (zero-based) month for the given *datetime* value, in UTC time."""
        return FunctionExpression("utcmonth", args)

    @classmethod
    def utcweek(cls, *args) -> FunctionExpression:
        """
        Returns the week number of the year for the given *datetime*, in UTC time.

        This function assumes Sunday-based weeks.
        Days before the first Sunday of the year are considered to be in week 0,
        the first Sunday of the year is the start of week 1,
        the second Sunday week 2, etc.
        """
        return FunctionExpression("utcweek", args)

    @classmethod
    def utchours(cls, *args) -> FunctionExpression:
        """Returns the hours component for the given *datetime* value, in UTC time."""
        return FunctionExpression("utchours", args)

    @classmethod
    def utcminutes(cls, *args) -> FunctionExpression:
        """Returns the minutes component for the given *datetime* value, in UTC time."""
        return FunctionExpression("utcminutes", args)

    @classmethod
    def utcseconds(cls, *args) -> FunctionExpression:
        """Returns the seconds component for the given *datetime* value, in UTC time."""
        return FunctionExpression("utcseconds", args)

    @classmethod
    def utcmilliseconds(cls, *args) -> FunctionExpression:
        """Returns the milliseconds component for the given *datetime* value, in UTC time."""
        return FunctionExpression("utcmilliseconds", args)

    @classmethod
    def utcOffset(cls, *args) -> FunctionExpression:
        """
        Returns a new `Date` instance that offsets the given *date* by the specified time `unit <https://vega.github.io/vega/docs/api/time/#time-units>`__ in UTC time.

        The optional *step* argument indicates the number of time unit steps to offset by (default 1).
        """
        return FunctionExpression("utcOffset", args)

    @classmethod
    def utcSequence(cls, *args) -> FunctionExpression:
        """
        Returns an array of `Date` instances from *start* (inclusive) to *stop* (exclusive), with each entry separated by the given time `unit <https://vega.github.io/vega/docs/api/time/#time-units>`__ in UTC time.

        The optional *step* argument indicates the number of time unit steps to take between each sequence entry (default 1).
        """
        return FunctionExpression("utcSequence", args)

    @classmethod
    def extent(cls, *args) -> FunctionExpression:
        """Returns a new `[min, max]` array with the minimum and maximum values of the input array, ignoring `null`, `undefined`, and `NaN` values."""
        return FunctionExpression("extent", args)

    @classmethod
    def clampRange(cls, *args) -> FunctionExpression:
        """
        Clamps a two-element *range* array in a span-preserving manner.

        If the span of the input *range* is less than `(max - min)` and an endpoint exceeds either the *min* or *max* value,
        the range is translated such that the span is preserved and one endpoint touches the boundary of the `[min, max]` range.
        If the span exceeds `(max - min)`, the range `[min, max]` is returned.
        """
        return FunctionExpression("clampRange", args)

    @classmethod
    def indexof(cls, *args) -> FunctionExpression:
        """Returns the first index of *value* in the input *array*, or the first index of *substring* in the input *string*."""
        return FunctionExpression("indexof", args)

    @classmethod
    def inrange(cls, *args) -> FunctionExpression:
        """Tests whether *value* lies within (or is equal to either) the first and last values of the *range* array."""
        return FunctionExpression("inrange", args)

    @classmethod
    def join(cls, *args) -> FunctionExpression:
        """Returns a new string by concatenating all of the elements of the input *array*, separated by commas or a specified *separator* string."""
        return FunctionExpression("join", args)

    @classmethod
    def lastindexof(cls, *args) -> FunctionExpression:
        """Returns the last index of *value* in the input *array*, or the last index of *substring* in the input *string*."""
        return FunctionExpression("lastindexof", args)

    @classmethod
    def length(cls, *args) -> FunctionExpression:
        """Returns the length of the input *array*, or the length of the input *string*."""
        return FunctionExpression("length", args)

    @classmethod
    def lerp(cls, *args) -> FunctionExpression:
        """
        Returns the linearly interpolated value between the first and last entries in the *array* for the provided interpolation *fraction* (typically between 0 and 1).

        For example, `lerp([0, 50], 0.5)` returns 25.
        """
        return FunctionExpression("lerp", args)

    @classmethod
    def peek(cls, *args) -> FunctionExpression:
        """
        Returns the last element in the input *array*.

        Similar to the built-in `Array.pop` method, except that it does not remove the last element.
        This method is a convenient shorthand for `array[array.length - 1]`.
        """
        return FunctionExpression("peek", args)

    @classmethod
    def pluck(cls, *args) -> FunctionExpression:
        """
        Retrieves the value for the specified *field* from a given *array* of objects.

        The input *field* string may include nested properties (e.g., `foo.bar.bz`).
        """
        return FunctionExpression("pluck", args)

    @classmethod
    def reverse(cls, *args) -> FunctionExpression:
        """
        Returns a new array with elements in a reverse order of the input *array*.

        The first array element becomes the last, and the last array element becomes the first.
        """
        return FunctionExpression("reverse", args)

    @classmethod
    def sequence(cls, *args) -> FunctionExpression:
        r"""
        Returns an array containing an arithmetic sequence of numbers.

        If *step* is omitted, it defaults to 1.
        If *start* is omitted, it defaults to 0.

        The *stop* value is exclusive; it is not included in the result.
        If *step* is positive, the last element is the largest `start + i * step` less than *stop*;
        if *step* is negative, the last element is the smallest `start + i * step` greater than *stop*.

        If the returned array would contain an infinite number of values, an empty range is returned.
        The arguments are not required to be integers.
        """
        return FunctionExpression("sequence", args)

    @classmethod
    def slice(cls, *args) -> FunctionExpression:
        """
        Returns a section of *array* between the *start* and *end* indices.

        If the *end* argument is negative, it is treated as an offset from the end of the array `length(array) + end`.
        """
        return FunctionExpression("slice", args)

    @classmethod
    def span(cls, *args) -> FunctionExpression:
        """
        Returns the span of *array*: the difference between the last and first elements, or `array[array.length-1] - array[0]`.

        Or if input is a string: a section of *string* between the *start* and *end* indices.
        If the *end* argument is negative, it is treated as an offset from the end of the string `length(string) + end`.
        """
        return FunctionExpression("span", args)

    @classmethod
    def lower(cls, *args) -> FunctionExpression:
        """Transforms *string* to lower-case letters."""
        return FunctionExpression("lower", args)

    @classmethod
    def pad(cls, *args) -> FunctionExpression:
        """
        Pads a *string* value with repeated instances of a *character* up to a specified *length*.

        If *character* is not specified, a space (' ') is used.
        By default, padding is added to the end of a string.
        An optional *align* parameter specifies if padding should be added to the `'left'` (beginning), `'center'`, or `'right'` (end) of the input string.
        """
        return FunctionExpression("pad", args)

    @classmethod
    def parseFloat(cls, *args) -> FunctionExpression:
        """
        Parses the input *string* to a floating-point value.

        Same as JavaScript's `parseFloat`.
        """
        return FunctionExpression("parseFloat", args)

    @classmethod
    def parseInt(cls, *args) -> FunctionExpression:
        """
        Parses the input *string* to an integer value.

        Same as JavaScript's `parseInt`.
        """
        return FunctionExpression("parseInt", args)

    @classmethod
    def replace(cls, *args) -> FunctionExpression:
        """
        Returns a new string with some or all matches of *pattern* replaced by a *replacement* string.

        The *pattern* can be a string or a regular expression.
        If *pattern* is a string, only the first instance will be replaced.
        Same as `JavaScript's String.replace <https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/String/replace>`__.
        """
        return FunctionExpression("replace", args)

    @classmethod
    def split(cls, *args) -> FunctionExpression:
        """
        Returns an array of tokens created by splitting the input *string* according to a provided *separator* pattern.

        The result can optionally be constrained to return at most *limit* tokens.
        """
        return FunctionExpression("split", args)

    @classmethod
    def substring(cls, *args) -> FunctionExpression:
        """Returns a section of *string* between the *start* and *end* indices."""
        return FunctionExpression("substring", args)

    @classmethod
    def trim(cls, *args) -> FunctionExpression:
        """Returns a trimmed string with preceding and trailing whitespace removed."""
        return FunctionExpression("trim", args)

    @classmethod
    def truncate(cls, *args) -> FunctionExpression:
        r"""
        Truncates an input *string* to a target *length*.

        The optional *align* argument indicates what part of the string should be truncated: `'left'` (the beginning), `'center'`, or `'right'` (the end).
        By default, the `'right'` end of the string is truncated.
        The optional *ellipsis* argument indicates the string to use to indicate truncated content;
        by default the ellipsis character `...` (`\\u2026`) is used.
        """
        return FunctionExpression("truncate", args)

    @classmethod
    def upper(cls, *args) -> FunctionExpression:
        """Transforms *string* to upper-case letters."""
        return FunctionExpression("upper", args)

    @classmethod
    def merge(cls, *args) -> FunctionExpression:
        """
        Merges the input objects *object1*, *object2*, etc into a new output object.

        Inputs are visited in sequential order, such that key values from later arguments can overwrite those from earlier arguments.

        Example: `merge({a:1, b:2}, {a:3}) -> {a:3, b:2}`.
        """
        return FunctionExpression("merge", args)

    @classmethod
    def dayFormat(cls, *args) -> FunctionExpression:
        """
        Formats a (0-6) *weekday* number as a full week day name, according to the current locale.

        For example: `dayFormat(0) -> "Sunday"`.
        """
        return FunctionExpression("dayFormat", args)

    @classmethod
    def dayAbbrevFormat(cls, *args) -> FunctionExpression:
        """
        Formats a (0-6) *weekday* number as an abbreviated week day name, according to the current locale.

        For example: `dayAbbrevFormat(0) -> "Sun"`.
        """
        return FunctionExpression("dayAbbrevFormat", args)

    @classmethod
    def format(cls, *args) -> FunctionExpression:
        """
        Formats a numeric *value* as a string.

        The *specifier* must be a valid `d3-format specifier <https://d3js.org/d3-format/>`__ (e.g., `format(value, ',.2f')`.
        """
        return FunctionExpression("format", args)

    @classmethod
    def monthFormat(cls, *args) -> FunctionExpression:
        """
        Formats a (zero-based) *month* number as a full month name, according to the current locale.

        For example: `monthFormat(0) -> "January"`.
        """
        return FunctionExpression("monthFormat", args)

    @classmethod
    def monthAbbrevFormat(cls, *args) -> FunctionExpression:
        """
        Formats a (zero-based) *month* number as an abbreviated month name, according to the current locale.

        For example: `monthAbbrevFormat(0) -> "Jan"`.
        """
        return FunctionExpression("monthAbbrevFormat", args)

    @classmethod
    def timeUnitSpecifier(cls, *args) -> FunctionExpression:
        """
        Returns a time format specifier string for the given time `unit <https://vega.github.io/vega/docs/api/time/#time-units>`__.

        The optional *specifiers* object provides a set of specifier sub-strings for customizing the format;
        for more, see the `timeUnitSpecifier API documentation <https://vega.github.io/vega/docs/api/time/#timeUnitSpecifier>`__.

        The resulting specifier string can then be used as input to the `timeFormat <https://vega.github.io/vega/docs/expressions/#timeFormat>`__ or
        `utcFormat <https://vega.github.io/vega/docs/expressions/#utcFormat>`__ functions, or as the *format* parameter of an axis or legend.

        For example: `timeFormat(date, timeUnitSpecifier('year'))` or `timeFormat(date, timeUnitSpecifier(['hours', 'minutes']))`.
        """
        return FunctionExpression("timeUnitSpecifier", args)

    @classmethod
    def timeFormat(cls, *args) -> FunctionExpression:
        """
        Formats a datetime *value* (either a `Date` object or timestamp) as a string, according to the local time.

        The *specifier* must be a valid `d3-time-format specifier <https://d3js.org/d3-time-format/>`__.
        For example: `timeFormat(timestamp, '%A')`.
        """
        return FunctionExpression("timeFormat", args)

    @classmethod
    def timeParse(cls, *args) -> FunctionExpression:
        """
        Parses a *string* value to a Date object, according to the local time.

        The *specifier* must be a valid `d3-time-format specifier <https://d3js.org/d3-time-format/>`__.
        For example: `timeParse('June 30, 2015', '%B %d, %Y')`.
        """
        return FunctionExpression("timeParse", args)

    @classmethod
    def utcFormat(cls, *args) -> FunctionExpression:
        """
        Formats a datetime *value* (either a `Date` object or timestamp) as a string, according to `UTC <https://en.wikipedia.org/wiki/Coordinated_Universal_Time>`__ time.

        The *specifier* must be a valid `d3-time-format specifier <https://d3js.org/d3-time-format/>`__.
        For example: `utcFormat(timestamp, '%A')`.
        """
        return FunctionExpression("utcFormat", args)

    @classmethod
    def utcParse(cls, *args) -> FunctionExpression:
        """
        Parses a *string* value to a Date object, according to `UTC <https://en.wikipedia.org/wiki/Coordinated_Universal_Time>`__ time.

        The *specifier* must be a valid `d3-time-format specifier <https://d3js.org/d3-time-format/>`__.
        For example: `utcParse('June 30, 2015', '%B %d, %Y')`.
        """
        return FunctionExpression("utcParse", args)

    @classmethod
    def regexp(cls, *args) -> FunctionExpression:
        """
        Creates a regular expression instance from an input *pattern* string and optional *flags*.

        Same as `JavaScript's `RegExp` <https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/RegExp>`__.
        """
        return FunctionExpression("regexp", args)

    @classmethod
    def test(cls, *args) -> FunctionExpression:
        r"""
        Evaluates a regular expression *regexp* against the input *string*, returning `true` if the string matches the pattern, `false` otherwise.

        For example: `test(\d{3}, "32-21-9483") -> true`.
        """
        return FunctionExpression("test", args)

    @classmethod
    def rgb(cls, *args) -> FunctionExpression:
        """
        Constructs a new `RGB <https://en.wikipedia.org/wiki/RGB_color_model>`__ color.

        If *r*, *g* and *b* are specified, these represent the channel values of the returned color; an *opacity* may also be specified.
        If a CSS Color Module Level 3 *specifier* string is specified, it is parsed and then converted to the RGB color space. Uses `d3-color's rgb function <https://d3js.org/d3-color#color_rgb>`__.
        """
        return FunctionExpression("rgb", args)

    @classmethod
    def hsl(cls, *args) -> FunctionExpression:
        """
        Constructs a new `HSL <https://en.wikipedia.org/wiki/HSL_and_HSV>`__ color.

        If *h*, *s* and *l* are specified, these represent the channel values of the returned color; an *opacity* may also be specified.
        If a CSS Color Module Level 3 *specifier* string is specified, it is parsed and then converted to the HSL color space.
        Uses `d3-color's hsl function <https://d3js.org/d3-color#hsl>`__.
        """
        return FunctionExpression("hsl", args)

    @classmethod
    def lab(cls, *args) -> FunctionExpression:
        """
        Constructs a new `CIE LAB <https://en.wikipedia.org/wiki/Lab_color_space#CIELAB>`__ color.

        If *l*, *a* and *b* are specified, these represent the channel values of the returned color; an *opacity* may also be specified.
        If a CSS Color Module Level 3 *specifier* string is specified, it is parsed and then converted to the LAB color space.
        Uses `d3-color's lab function <https://d3js.org/d3-color#lab>`__.
        """
        return FunctionExpression("lab", args)

    @classmethod
    def hcl(cls, *args) -> FunctionExpression:
        """
        Constructs a new `HCL <https://en.wikipedia.org/wiki/Lab_color_space#CIELAB>`__ (hue, chroma, luminance) color.

        If *h*, *c* and *l* are specified, these represent the channel values of the returned color; an *opacity* may also be specified.
        If a CSS Color Module Level 3 *specifier* string is specified, it is parsed and then converted to the HCL color space.
        Uses `d3-color's hcl function <https://d3js.org/d3-color#hcl>`__.
        """
        return FunctionExpression("hcl", args)

    @classmethod
    def luminance(cls, *args) -> FunctionExpression:
        """
        Returns the luminance for the given color *specifier* (compatible with `d3-color's rgb function <https://d3js.org/d3-color#rgb)>`__.

        The luminance is calculated according to the `W3C Web Content Accessibility Guidelines <https://www.w3.org/TR/2008/REC-WCAG20-20081211/#relativeluminancedef>`__.
        """
        return FunctionExpression("luminance", args)

    @classmethod
    def contrast(cls, *args) -> FunctionExpression:
        """
        Returns the contrast ratio between the input color specifiers as a float between 1 and 21.

        The contrast is calculated according to the `W3C Web Content Accessibility Guidelines <https://www.w3.org/TR/2008/REC-WCAG20-20081211/#contrast-ratiodef>`__.
        """
        return FunctionExpression("contrast", args)

    @classmethod
    def item(cls, *args) -> FunctionExpression:
        """Returns the current scenegraph item that is the target of the event."""
        return FunctionExpression("item", args)

    @classmethod
    def group(cls, *args) -> FunctionExpression:
        """
        Returns the scenegraph group mark item in which the current event has occurred.

        If no arguments are provided, the immediate parent group is returned.
        If a group name is provided, the matching ancestor group item is returned.
        """
        return FunctionExpression("group", args)

    @classmethod
    def xy(cls, *args) -> FunctionExpression:
        """
        Returns the x- and y-coordinates for the current event as a two-element array.

        If no arguments are provided, the top-level coordinate space of the view is used.
        If a scenegraph *item* (or string group name) is provided, the coordinate space of the group item is used.
        """
        return FunctionExpression("xy", args)

    @classmethod
    def x(cls, *args) -> FunctionExpression:
        """
        Returns the x coordinate for the current event.

        If no arguments are provided, the top-level coordinate space of the view is used.
        If a scenegraph *item* (or string group name) is provided, the coordinate space of the group item is used.
        """
        return FunctionExpression("x", args)

    @classmethod
    def y(cls, *args) -> FunctionExpression:
        """
        Returns the y coordinate for the current event.

        If no arguments are provided, the top-level coordinate space of the view is used.
        If a scenegraph *item* (or string group name) is provided, the coordinate space of the group item is used.
        """
        return FunctionExpression("y", args)

    @classmethod
    def pinchDistance(cls, *args) -> FunctionExpression:
        """Returns the pixel distance between the first two touch points of a multi-touch event."""
        return FunctionExpression("pinchDistance", args)

    @classmethod
    def pinchAngle(cls, *args) -> FunctionExpression:
        """Returns the angle of the line connecting the first two touch points of a multi-touch event."""
        return FunctionExpression("pinchAngle", args)

    @classmethod
    def inScope(cls, *args) -> FunctionExpression:
        """Returns true if the given scenegraph *item* is a descendant of the group mark in which the event handler was defined, false otherwise."""
        return FunctionExpression("inScope", args)

    @classmethod
    def data(cls, *args) -> FunctionExpression:
        """
        Returns the array of data objects for the Vega data set with the given *name*.

        If the data set is not found, returns an empty array.
        """
        return FunctionExpression("data", args)

    @classmethod
    def indata(cls, *args) -> FunctionExpression:
        """
        Tests if the data set with a given *name* contains a datum with a *field* value that matches the input *value*.

        For example: `indata('table', 'category', value)`.
        """
        return FunctionExpression("indata", args)

    @classmethod
    def scale(cls, *args) -> FunctionExpression:
        """
        Applies the named scale transform (or projection) to the specified *value*.

        The optional *group* argument takes a scenegraph group mark item to indicate the specific scope in which to look up the scale or projection.
        """
        return FunctionExpression("scale", args)

    @classmethod
    def invert(cls, *args) -> FunctionExpression:
        """
        Inverts the named scale transform (or projection) for the specified *value*.

        The optional *group* argument takes a scenegraph group mark item to indicate the specific scope in which to look up the scale or projection.
        """
        return FunctionExpression("invert", args)

    @classmethod
    def copy(cls, *args) -> FunctionExpression:  # type: ignore[override]
        """
        Returns a copy (a new cloned instance) of the named scale transform of projection, or `undefined` if no scale or projection is found.

        The optional *group* argument takes a scenegraph group mark item to indicate the specific scope in which to look up the scale or projection.
        """
        # error: Signature of "copy" incompatible with supertype "SchemaBase"  [override]
        # note:      def copy(self, deep: bool | Iterable[Any] = ..., ignore: list[str] | None = ...) -> expr
        # NOTE: Not relevant as `expr() -> ExprRef`
        # this method is only accesible via `expr.copy()`
        return FunctionExpression("copy", args)

    @classmethod
    def domain(cls, *args) -> FunctionExpression:
        """
        Returns the scale domain array for the named scale transform, or an empty array if the scale is not found.

        The optional *group* argument takes a scenegraph group mark item to indicate the specific scope in which to look up the scale.
        """
        return FunctionExpression("domain", args)

    @classmethod
    def range(cls, *args) -> FunctionExpression:
        """
        Returns the scale range array for the named scale transform, or an empty array if the scale is not found.

        The optional *group* argument takes a scenegraph group mark item to indicate the specific scope in which to look up the scale.
        """
        return FunctionExpression("range", args)

    @classmethod
    def bandwidth(cls, *args) -> FunctionExpression:
        """
        Returns the current band width for the named band scale transform, or zero if the scale is not found or is not a band scale.

        The optional *group* argument takes a scenegraph group mark item to indicate the specific scope in which to look up the scale.
        """
        return FunctionExpression("bandwidth", args)

    @classmethod
    def bandspace(cls, *args) -> FunctionExpression:
        """
        Returns the number of steps needed within a band scale, based on the *count* of domain elements and the inner and outer padding values.

        While normally calculated within the scale itself, this function can be helpful for determining the size of a chart's layout.
        """
        return FunctionExpression("bandspace", args)

    @classmethod
    def gradient(cls, *args) -> FunctionExpression:
        """
        Returns a linear color gradient for the *scale* (whose range must be a `continuous color scheme <https://vega.github.io/vega/docs/schemes>`__ and starting and ending points *p0* and *p1*, each an `[x, y]` array.

        The points *p0* and *p1* should be expressed in normalized coordinates in the domain `[0, 1]`, relative to the bounds of the item being colored.

        If unspecified, *p0* defaults to `[0, 0]` and *p1* defaults to `[1, 0]`, for a horizontal gradient that spans the full bounds of an item.
        The optional *count* argument indicates a desired target number of sample points to take from the color scale.
        """
        return FunctionExpression("gradient", args)

    @classmethod
    def panLinear(cls, *args) -> FunctionExpression:
        """
        Given a linear scale *domain* array with numeric or datetime values, returns a new two-element domain array that is the result of panning the domain by a fractional *delta*.

        The *delta* value represents fractional units of the scale range; for example, `0.5` indicates panning the scale domain to the right by half the scale range.
        """
        return FunctionExpression("panLinear", args)

    @classmethod
    def panLog(cls, *args) -> FunctionExpression:
        """
        Given a log scale *domain* array with numeric or datetime values, returns a new two-element domain array that is the result of panning the domain by a fractional *delta*.

        The *delta* value represents fractional units of the scale range; for example, `0.5` indicates panning the scale domain to the right by half the scale range.
        """
        return FunctionExpression("panLog", args)

    @classmethod
    def panPow(cls, *args) -> FunctionExpression:
        """
        Given a power scale *domain* array with numeric or datetime values and the given *exponent*, returns a new two-element domain array that is the result of panning the domain by a fractional *delta*.

        The *delta* value represents fractional units of the scale range; for example, `0.5` indicates panning the scale domain to the right by half the scale range.
        """
        return FunctionExpression("panPow", args)

    @classmethod
    def panSymlog(cls, *args) -> FunctionExpression:
        """
        Given a symmetric log scale *domain* array with numeric or datetime values parameterized by the given *constant*, returns a new two-element domain array that is the result of panning the domain by a fractional *delta*.

        The *delta* value represents fractional units of the scale range; for example, `0.5` indicates panning the scale domain to the right by half the scale range.
        """
        return FunctionExpression("panSymlog", args)

    @classmethod
    def zoomLinear(cls, *args) -> FunctionExpression:
        """
        Given a linear scale *domain* array with numeric or datetime values, returns a new two-element domain array that is the result of zooming the domain by a *scaleFactor*, centered at the provided fractional *anchor*.

        The *anchor* value represents the zoom position in terms of fractional units of the scale range; for example, `0.5` indicates a zoom centered on the mid-point of the scale range.
        """
        return FunctionExpression("zoomLinear", args)

    @classmethod
    def zoomLog(cls, *args) -> FunctionExpression:
        """
        Given a log scale *domain* array with numeric or datetime values, returns a new two-element domain array that is the result of zooming the domain by a *scaleFactor*, centered at the provided fractional *anchor*.

        The *anchor* value represents the zoom position in terms of fractional units of the scale range; for example, `0.5` indicates a zoom centered on the mid-point of the scale range.
        """
        return FunctionExpression("zoomLog", args)

    @classmethod
    def zoomPow(cls, *args) -> FunctionExpression:
        """
        Given a power scale *domain* array with numeric or datetime values and the given *exponent*, returns a new two-element domain array that is the result of zooming the domain by a *scaleFactor*, centered at the provided fractional *anchor*.

        The *anchor* value represents the zoom position in terms of fractional units of the scale range; for example, `0.5` indicates a zoom centered on the mid-point of the scale range.
        """
        return FunctionExpression("zoomPow", args)

    @classmethod
    def zoomSymlog(cls, *args) -> FunctionExpression:
        """
        Given a symmetric log scale *domain* array with numeric or datetime values parameterized by the given *constant*, returns a new two-element domain array that is the result of zooming the domain by a *scaleFactor*, centered at the provided fractional *anchor*.

        The *anchor* value represents the zoom position in terms of fractional units of the scale range; for example, `0.5` indicates a zoom centered on the mid-point of the scale range.
        """
        return FunctionExpression("zoomSymlog", args)

    @classmethod
    def geoArea(cls, *args) -> FunctionExpression:
        """
        Returns the projected planar area (typically in square pixels) of a GeoJSON *feature* according to the named *projection*.

        If the *projection* argument is `null`, computes the spherical area in steradians using unprojected longitude, latitude coordinates.
        The optional *group* argument takes a scenegraph group mark item to indicate the specific scope in which to look up the projection.
        Uses d3-geo's `geoArea <https://d3js.org/d3-geo#geoArea>`__ and `path.area <https://d3js.org/d3-geo#path_area>`__ methods.
        """
        return FunctionExpression("geoArea", args)

    @classmethod
    def geoBounds(cls, *args) -> FunctionExpression:
        """
        Returns the projected planar bounding box (typically in pixels) for the specified GeoJSON *feature*, according to the named *projection*.

        The bounding box is represented by a two-dimensional array: `[[x0, y0], [x1, y1]]`,
        where *x0* is the minimum x-coordinate, *y0* is the minimum y-coordinate,
        *x1* is the maximum x-coordinate, and *y1* is the maximum y-coordinate.

        If the *projection* argument is `null`, computes the spherical bounding box using unprojected longitude, latitude coordinates.
        The optional *group* argument takes a scenegraph group mark item to indicate the specific scope in which to look up the projection.
        Uses d3-geo's `geoBounds <https://d3js.org/d3-geo#geoBounds>`__ and `path.bounds <https://d3js.org/d3-geo#path_bounds>`__ methods.
        """
        return FunctionExpression("geoBounds", args)

    @classmethod
    def geoCentroid(cls, *args) -> FunctionExpression:
        """
        Returns the projected planar centroid (typically in pixels) for the specified GeoJSON *feature*, according to the named *projection*.

        If the *projection* argument is `null`, computes the spherical centroid using unprojected longitude, latitude coordinates.
        The optional *group* argument takes a scenegraph group mark item to indicate the specific scope in which to look up the projection.
        Uses d3-geo's `geoCentroid <https://d3js.org/d3-geo#geoCentroid>`__ and `path.centroid <https://d3js.org/d3-geo#path_centroid>`__ methods.
        """
        return FunctionExpression("geoCentroid", args)

    @classmethod
    def treePath(cls, *args) -> FunctionExpression:
        """
        For the hierarchy data set with the given *name*, returns the shortest path through from the *source* node id to the *target* node id.

        The path starts at the *source* node, ascends to the least common ancestor of the *source* node and the *target* node, and then descends to the *target* node.
        """
        return FunctionExpression("treePath", args)

    @classmethod
    def treeAncestors(cls, *args) -> FunctionExpression:
        """For the hierarchy data set with the given *name*, returns the array of ancestors nodes, starting with the input *node*, then followed by each parent up to the root."""
        return FunctionExpression("treeAncestors", args)

    @classmethod
    def containerSize(cls, *args) -> FunctionExpression:
        """
        Returns the current CSS box size (`[el.clientWidth, el.clientHeight]`) of the parent DOM element that contains the Vega view.

        If there is no container element, returns `[undefined, undefined]`.
        """
        return FunctionExpression("containerSize", args)

    @classmethod
    def screen(cls, *args) -> FunctionExpression:
        """Returns the `window.screen <https://developer.mozilla.org/en-US/docs/Web/API/Window/screen>`__ object, or `{}` if Vega is not running in a browser environment."""
        return FunctionExpression("screen", args)

    @classmethod
    def windowSize(cls, *args) -> FunctionExpression:
        """Returns the current window size (`[window.innerWidth, window.innerHeight]`) or `[undefined, undefined]` if Vega is not running in a browser environment."""
        return FunctionExpression("windowSize", args)

    @classmethod
    def warn(cls, *args) -> FunctionExpression:
        """
        Logs a warning message and returns the last argument.

        For the message to appear in the console, the visualization view must have the appropriate logging level set.
        """
        return FunctionExpression("warn", args)

    @classmethod
    def info(cls, *args) -> FunctionExpression:
        """
        Logs an informative message and returns the last argument.

        For the message to appear in the console, the visualization view must have the appropriate logging level set.
        """
        return FunctionExpression("info", args)

    @classmethod
    def debug(cls, *args) -> FunctionExpression:
        """
        Logs a debugging message and returns the last argument.

        For the message to appear in the console, the visualization view must have the appropriate logging level set.
        """
        return FunctionExpression("debug", args)


_ExprType = expr
# NOTE: Compatibility alias for previous type of `alt.expr`.
# `_ExprType` was not referenced in any internal imports/tests.



from __future__ import annotations

import hashlib
import json
import random
import sys
from functools import partial
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Literal,
    MutableMapping,
    Protocol,
    Sequence,
    TypedDict,
    TypeVar,
    Union,
    overload,
    runtime_checkable,
)
from typing_extensions import Concatenate, ParamSpec, TypeAlias

import narwhals.stable.v1 as nw
from narwhals.dependencies import is_pandas_dataframe as _is_pandas_dataframe
from narwhals.typing import IntoDataFrame

from ._importers import import_pyarrow_interchange
from .core import (
    DataFrameLike,
    sanitize_geo_interface,
    sanitize_narwhals_dataframe,
    sanitize_pandas_dataframe,
    to_eager_narwhals_dataframe,
)
from .plugin_registry import PluginRegistry

if sys.version_info >= (3, 13):
    from typing import TypeIs
else:
    from typing_extensions import TypeIs

if TYPE_CHECKING:
    import pandas as pd
    import pyarrow as pa


@runtime_checkable
class SupportsGeoInterface(Protocol):
    __geo_interface__: MutableMapping


DataType: TypeAlias = Union[
    Dict[Any, Any], IntoDataFrame, SupportsGeoInterface, DataFrameLike
]

TDataType = TypeVar("TDataType", bound=DataType)
TIntoDataFrame = TypeVar("TIntoDataFrame", bound=IntoDataFrame)

VegaLiteDataDict: TypeAlias = Dict[
    str, Union[str, Dict[Any, Any], List[Dict[Any, Any]]]
]
ToValuesReturnType: TypeAlias = Dict[str, Union[Dict[Any, Any], List[Dict[Any, Any]]]]
SampleReturnType = Union[IntoDataFrame, Dict[str, Sequence], None]


def is_data_type(obj: Any) -> TypeIs[DataType]:
    return _is_pandas_dataframe(obj) or isinstance(
        obj, (dict, DataFrameLike, SupportsGeoInterface, nw.DataFrame)
    )


# ==============================================================================
# Data transformer registry
#
# A data transformer is a callable that takes a supported data type and returns
# a transformed dictionary version of it which is compatible with the VegaLite schema.
# The dict objects will be the Data portion of the VegaLite schema.
#
# Renderers only deal with the dict form of a
# VegaLite spec, after the Data model has been put into a schema compliant
# form.
# ==============================================================================

P = ParamSpec("P")
# NOTE: `Any` required due to the complexity of existing signatures imported in `altair.vegalite.v5.data.py`
R = TypeVar("R", VegaLiteDataDict, Any)
DataTransformerType = Callable[Concatenate[DataType, P], R]


class DataTransformerRegistry(PluginRegistry[DataTransformerType, R]):
    _global_settings = {"consolidate_datasets": True}

    @property
    def consolidate_datasets(self) -> bool:
        return self._global_settings["consolidate_datasets"]

    @consolidate_datasets.setter
    def consolidate_datasets(self, value: bool) -> None:
        self._global_settings["consolidate_datasets"] = value


# ==============================================================================
class MaxRowsError(Exception):
    """Raised when a data model has too many rows."""


@overload
def limit_rows(data: None = ..., max_rows: int | None = ...) -> partial: ...
@overload
def limit_rows(data: DataType, max_rows: int | None = ...) -> DataType: ...
def limit_rows(
    data: DataType | None = None, max_rows: int | None = 5000
) -> partial | DataType:
    """
    Raise MaxRowsError if the data model has more than max_rows.

    If max_rows is None, then do not perform any check.
    """
    if data is None:
        return partial(limit_rows, max_rows=max_rows)
    check_data_type(data)

    def raise_max_rows_error():
        msg = (
            "The number of rows in your dataset is greater "
            f"than the maximum allowed ({max_rows}).\n\n"
            "Try enabling the VegaFusion data transformer which "
            "raises this limit by pre-evaluating data\n"
            "transformations in Python.\n"
            "    >> import altair as alt\n"
            '    >> alt.data_transformers.enable("vegafusion")\n\n'
            "Or, see https://altair-viz.github.io/user_guide/large_datasets.html "
            "for additional information\n"
            "on how to plot large datasets."
        )
        raise MaxRowsError(msg)

    if isinstance(data, SupportsGeoInterface):
        if data.__geo_interface__["type"] == "FeatureCollection":
            values = data.__geo_interface__["features"]
        else:
            values = data.__geo_interface__
    elif isinstance(data, dict):
        if "values" in data:
            values = data["values"]
        else:
            return data
    else:
        data = to_eager_narwhals_dataframe(data)
        values = data

    if max_rows is not None and len(values) > max_rows:
        raise_max_rows_error()

    return data


@overload
def sample(
    data: None = ..., n: int | None = ..., frac: float | None = ...
) -> partial: ...
@overload
def sample(
    data: TIntoDataFrame, n: int | None = ..., frac: float | None = ...
) -> TIntoDataFrame: ...
@overload
def sample(
    data: DataType, n: int | None = ..., frac: float | None = ...
) -> SampleReturnType: ...
def sample(
    data: DataType | None = None,
    n: int | None = None,
    frac: float | None = None,
) -> partial | SampleReturnType:
    """Reduce the size of the data model by sampling without replacement."""
    if data is None:
        return partial(sample, n=n, frac=frac)
    check_data_type(data)
    if _is_pandas_dataframe(data):
        return data.sample(n=n, frac=frac)
    elif isinstance(data, dict):
        if "values" in data:
            values = data["values"]
            if not n:
                if frac is None:
                    msg = "frac cannot be None if n is None and data is a dictionary"
                    raise ValueError(msg)
                n = int(frac * len(values))
            values = random.sample(values, n)
            return {"values": values}
        else:
            # Maybe this should raise an error or return something useful?
            return None
    data = nw.from_native(data, eager_only=True)
    if not n:
        if frac is None:
            msg = "frac cannot be None if n is None with this data input type"
            raise ValueError(msg)
        n = int(frac * len(data))
    indices = random.sample(range(len(data)), n)
    return nw.to_native(data[indices])


_FormatType = Literal["csv", "json"]


class _FormatDict(TypedDict):
    type: _FormatType


class _ToFormatReturnUrlDict(TypedDict):
    url: str
    format: _FormatDict


@overload
def to_json(
    data: None = ...,
    prefix: str = ...,
    extension: str = ...,
    filename: str = ...,
    urlpath: str = ...,
) -> partial: ...


@overload
def to_json(
    data: DataType,
    prefix: str = ...,
    extension: str = ...,
    filename: str = ...,
    urlpath: str = ...,
) -> _ToFormatReturnUrlDict: ...


def to_json(
    data: DataType | None = None,
    prefix: str = "altair-data",
    extension: str = "json",
    filename: str = "{prefix}-{hash}.{extension}",
    urlpath: str = "",
) -> partial | _ToFormatReturnUrlDict:
    """Write the data model to a .json file and return a url based data model."""
    kwds = _to_text_kwds(prefix, extension, filename, urlpath)
    if data is None:
        return partial(to_json, **kwds)
    else:
        data_str = _data_to_json_string(data)
        return _to_text(data_str, **kwds, format=_FormatDict(type="json"))


@overload
def to_csv(
    data: None = ...,
    prefix: str = ...,
    extension: str = ...,
    filename: str = ...,
    urlpath: str = ...,
) -> partial: ...


@overload
def to_csv(
    data: dict | pd.DataFrame | DataFrameLike,
    prefix: str = ...,
    extension: str = ...,
    filename: str = ...,
    urlpath: str = ...,
) -> _ToFormatReturnUrlDict: ...


def to_csv(
    data: dict | pd.DataFrame | DataFrameLike | None = None,
    prefix: str = "altair-data",
    extension: str = "csv",
    filename: str = "{prefix}-{hash}.{extension}",
    urlpath: str = "",
) -> partial | _ToFormatReturnUrlDict:
    """Write the data model to a .csv file and return a url based data model."""
    kwds = _to_text_kwds(prefix, extension, filename, urlpath)
    if data is None:
        return partial(to_csv, **kwds)
    else:
        data_str = _data_to_csv_string(data)
        return _to_text(data_str, **kwds, format=_FormatDict(type="csv"))


def _to_text(
    data: str,
    prefix: str,
    extension: str,
    filename: str,
    urlpath: str,
    format: _FormatDict,
) -> _ToFormatReturnUrlDict:
    data_hash = _compute_data_hash(data)
    filename = filename.format(prefix=prefix, hash=data_hash, extension=extension)
    Path(filename).write_text(data, encoding="utf-8")
    url = str(Path(urlpath, filename))
    return _ToFormatReturnUrlDict({"url": url, "format": format})


def _to_text_kwds(prefix: str, extension: str, filename: str, urlpath: str, /) -> dict[str, str]:  # fmt: skip
    return {"prefix": prefix, "extension": extension, "filename": filename, "urlpath": urlpath}  # fmt: skip


def to_values(data: DataType) -> ToValuesReturnType:
    """Replace a DataFrame by a data model with values."""
    check_data_type(data)
    # `strict=False` passes `data` through as-is if it is not a Narwhals object.
    data_native = nw.to_native(data, strict=False)
    if isinstance(data_native, SupportsGeoInterface):
        if _is_pandas_dataframe(data_native):
            data_native = sanitize_pandas_dataframe(data_native)
        # Maybe the type could be further clarified here that it is
        # SupportGeoInterface and then the ignore statement is not needed?
        data_sanitized = sanitize_geo_interface(data_native.__geo_interface__)
        return {"values": data_sanitized}
    elif _is_pandas_dataframe(data_native):
        data_native = sanitize_pandas_dataframe(data_native)
        return {"values": data_native.to_dict(orient="records")}
    elif isinstance(data_native, dict):
        if "values" not in data_native:
            msg = "values expected in data dict, but not present."
            raise KeyError(msg)
        return data_native
    elif isinstance(data, nw.DataFrame):
        data = sanitize_narwhals_dataframe(data)
        return {"values": data.rows(named=True)}
    else:
        # Should never reach this state as tested by check_data_type
        msg = f"Unrecognized data type: {type(data)}"
        raise ValueError(msg)


def check_data_type(data: DataType) -> None:
    if not is_data_type(data):
        msg = f"Expected dict, DataFrame or a __geo_interface__ attribute, got: {type(data)}"
        raise TypeError(msg)


# ==============================================================================
# Private utilities
# ==============================================================================
def _compute_data_hash(data_str: str) -> str:
    return hashlib.sha256(data_str.encode()).hexdigest()[:32]


def _data_to_json_string(data: DataType) -> str:
    """Return a JSON string representation of the input data."""
    check_data_type(data)
    # `strict=False` passes `data` through as-is if it is not a Narwhals object.
    data_native = nw.to_native(data, strict=False)
    if isinstance(data_native, SupportsGeoInterface):
        if _is_pandas_dataframe(data_native):
            data_native = sanitize_pandas_dataframe(data_native)
        data_native = sanitize_geo_interface(data_native.__geo_interface__)
        return json.dumps(data_native)
    elif _is_pandas_dataframe(data_native):
        data = sanitize_pandas_dataframe(data_native)
        return data_native.to_json(orient="records", double_precision=15)
    elif isinstance(data_native, dict):
        if "values" not in data_native:
            msg = "values expected in data dict, but not present."
            raise KeyError(msg)
        return json.dumps(data_native["values"], sort_keys=True)
    elif isinstance(data, nw.DataFrame):
        return json.dumps(data.rows(named=True))
    else:
        msg = "to_json only works with data expressed as " "a DataFrame or as a dict"
        raise NotImplementedError(msg)


def _data_to_csv_string(data: dict | pd.DataFrame | DataFrameLike) -> str:
    """Return a CSV string representation of the input data."""
    check_data_type(data)
    if isinstance(data, SupportsGeoInterface):
        msg = (
            f"to_csv does not yet work with data that "
            f"is of type {type(SupportsGeoInterface).__name__!r}.\n"
            f"See https://github.com/vega/altair/issues/3441"
        )
        raise NotImplementedError(msg)
    elif _is_pandas_dataframe(data):
        data = sanitize_pandas_dataframe(data)
        return data.to_csv(index=False)
    elif isinstance(data, dict):
        if "values" not in data:
            msg = "values expected in data dict, but not present"
            raise KeyError(msg)
        try:
            import pandas as pd
        except ImportError as exc:
            msg = "pandas is required to convert a dict to a CSV string"
            raise ImportError(msg) from exc
        return pd.DataFrame.from_dict(data["values"]).to_csv(index=False)
    elif isinstance(data, DataFrameLike):
        # experimental interchange dataframe support
        import pyarrow as pa
        import pyarrow.csv as pa_csv

        pa_table = arrow_table_from_dfi_dataframe(data)
        csv_buffer = pa.BufferOutputStream()
        pa_csv.write_csv(pa_table, csv_buffer)
        return csv_buffer.getvalue().to_pybytes().decode()
    else:
        msg = "to_csv only works with data expressed as " "a DataFrame or as a dict"
        raise NotImplementedError(msg)


def arrow_table_from_dfi_dataframe(dfi_df: DataFrameLike) -> pa.Table:
    """Convert a DataFrame Interchange Protocol compatible object to an Arrow Table."""
    import pyarrow as pa

    # First check if the dataframe object has a method to convert to arrow.
    # Give this preference over the pyarrow from_dataframe function since the object
    # has more control over the conversion, and may have broader compatibility.
    # This is the case for Polars, which supports Date32 columns in direct conversion
    # while pyarrow does not yet support this type in from_dataframe
    for convert_method_name in ("arrow", "to_arrow", "to_arrow_table", "to_pyarrow"):
        convert_method = getattr(dfi_df, convert_method_name, None)
        if callable(convert_method):
            result = convert_method()
            if isinstance(result, pa.Table):
                return result

    pi = import_pyarrow_interchange()
    return pi.from_dataframe(dfi_df)



"""Public types to ease integrating with `altair`."""

from __future__ import annotations

__all__ = [
    "ChannelAngle",
    "ChannelColor",
    "ChannelColumn",
    "ChannelDescription",
    "ChannelDetail",
    "ChannelFacet",
    "ChannelFill",
    "ChannelFillOpacity",
    "ChannelHref",
    "ChannelKey",
    "ChannelLatitude",
    "ChannelLatitude2",
    "ChannelLongitude",
    "ChannelLongitude2",
    "ChannelOpacity",
    "ChannelOrder",
    "ChannelRadius",
    "ChannelRadius2",
    "ChannelRow",
    "ChannelShape",
    "ChannelSize",
    "ChannelStroke",
    "ChannelStrokeDash",
    "ChannelStrokeOpacity",
    "ChannelStrokeWidth",
    "ChannelText",
    "ChannelTheta",
    "ChannelTheta2",
    "ChannelTooltip",
    "ChannelUrl",
    "ChannelX",
    "ChannelX2",
    "ChannelXError",
    "ChannelXError2",
    "ChannelXOffset",
    "ChannelY",
    "ChannelY2",
    "ChannelYError",
    "ChannelYError2",
    "ChannelYOffset",
    "ChartType",
    "EncodeKwds",
    "Optional",
    "is_chart_type",
]

from altair.utils.schemapi import Optional
from altair.vegalite.v5.api import ChartType, is_chart_type
from altair.vegalite.v5.schema.channels import (
    ChannelAngle,
    ChannelColor,
    ChannelColumn,
    ChannelDescription,
    ChannelDetail,
    ChannelFacet,
    ChannelFill,
    ChannelFillOpacity,
    ChannelHref,
    ChannelKey,
    ChannelLatitude,
    ChannelLatitude2,
    ChannelLongitude,
    ChannelLongitude2,
    ChannelOpacity,
    ChannelOrder,
    ChannelRadius,
    ChannelRadius2,
    ChannelRow,
    ChannelShape,
    ChannelSize,
    ChannelStroke,
    ChannelStrokeDash,
    ChannelStrokeOpacity,
    ChannelStrokeWidth,
    ChannelText,
    ChannelTheta,
    ChannelTheta2,
    ChannelTooltip,
    ChannelUrl,
    ChannelX,
    ChannelX2,
    ChannelXError,
    ChannelXError2,
    ChannelXOffset,
    ChannelY,
    ChannelY2,
    ChannelYError,
    ChannelYError2,
    ChannelYOffset,
    EncodeKwds,
)



from __future__ import annotations

import webbrowser
from http.server import BaseHTTPRequestHandler, HTTPServer
from typing import Iterable


def open_html_in_browser(
    html: str | bytes,
    using: str | Iterable[str] | None = None,
    port: int | None = None,
) -> None:
    """
    Display an html document in a web browser without creating a temp file.

    Instantiates a simple http server and uses the webbrowser module to
    open the server's URL

    Parameters
    ----------
    html: str
        HTML string to display
    using: str or iterable of str
        Name of the web browser to open (e.g. "chrome", "firefox", etc.).
        If an iterable, choose the first browser available on the system.
        If none, choose the system default browser.
    port: int
        Port to use. Defaults to a random port
    """
    # Encode html to bytes
    html_bytes = html.encode("utf8") if isinstance(html, str) else html

    browser = None

    if using is None:
        browser = webbrowser.get(None)
    else:
        # normalize using to an iterable
        if isinstance(using, str):
            using = [using]

        for browser_key in using:
            try:
                browser = webbrowser.get(browser_key)
                if browser is not None:
                    break
            except webbrowser.Error:
                pass

        if browser is None:
            raise ValueError("Failed to locate a browser with name in " + str(using))

    class OneShotRequestHandler(BaseHTTPRequestHandler):
        def do_GET(self) -> None:
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()

            bufferSize = 1024 * 1024
            for i in range(0, len(html_bytes), bufferSize):
                self.wfile.write(html_bytes[i : i + bufferSize])

        def log_message(self, format, *args):
            # Silence stderr logging
            pass

    # Use specified port if provided, otherwise choose a random port (port value of 0)
    server = HTTPServer(
        ("127.0.0.1", port if port is not None else 0), OneShotRequestHandler
    )
    browser.open(f"http://127.0.0.1:{server.server_port}")
    server.handle_request()



from altair.utils.display import (
    DefaultRendererReturnType,
    Displayable,
    HTMLRenderer,
    RendererRegistry,
    default_renderer_base,
    json_renderer_base,
)

__all__ = (
    "DefaultRendererReturnType",
    "Displayable",
    "HTMLRenderer",
    "RendererRegistry",
    "default_renderer_base",
    "json_renderer_base",
)



# DataFrame Interchange Protocol Types
# Copied from https://data-apis.org/dataframe-protocol/latest/API.html,
# changed ABCs to Protocols, and subset the type hints to only those that are
# relevant for Altair.
#
# These classes are only for use in type signatures
from __future__ import annotations

import enum
from typing import Any, Iterable, Protocol


class DtypeKind(enum.IntEnum):
    """
    Integer enum for data types.

    Attributes
    ----------
    INT : int
        Matches to signed integer data type.
    UINT : int
        Matches to unsigned integer data type.
    FLOAT : int
        Matches to floating point data type.
    BOOL : int
        Matches to boolean data type.
    STRING : int
        Matches to string data type (UTF-8 encoded).
    DATETIME : int
        Matches to datetime data type.
    CATEGORICAL : int
        Matches to categorical data type.
    """

    INT = 0
    UINT = 1
    FLOAT = 2
    BOOL = 20
    STRING = 21  # UTF-8
    DATETIME = 22
    CATEGORICAL = 23


# Type hint of first element would actually be DtypeKind but can't use that
# as other libraries won't use an instance of our own Enum in this module but have
# their own. Type checkers will raise an error on that even though the enums
# are identical.
class Column(Protocol):
    @property
    def dtype(self) -> tuple[Any, int, str, str]:
        """
        Dtype description as a tuple ``(kind, bit-width, format string, endianness)``.

        Bit-width : the number of bits as an integer
        Format string : data type description format string in Apache Arrow C
                        Data Interface format.
        Endianness : current only native endianness (``=``) is supported

        Notes
        -----
            - Kind specifiers are aligned with DLPack where possible (hence the
              jump to 20, leave enough room for future extension)
            - Masks must be specified as boolean with either bit width 1 (for bit
              masks) or 8 (for byte masks).
            - Dtype width in bits was preferred over bytes
            - Endianness isn't too useful, but included now in case in the future
              we need to support non-native endianness
            - Went with Apache Arrow format strings over NumPy format strings
              because they're more complete from a dataframe perspective
            - Format strings are mostly useful for datetime specification, and
              for categoricals.
            - For categoricals, the format string describes the type of the
              categorical in the data buffer. In case of a separate encoding of
              the categorical (e.g. an integer to string mapping), this can
              be derived from ``self.describe_categorical``.
            - Data types not included: complex, Arrow-style null, binary, decimal,
              and nested (list, struct, map, union) dtypes.
        """

    # Have to use a generic Any return type as not all libraries who implement
    # the dataframe interchange protocol implement the TypedDict that is usually
    # returned here in the same way. As TypedDicts are invariant, even a slight change
    # will lead to an error by a type checker. See PR in which this code was added
    # for details.
    @property
    def describe_categorical(self) -> Any:
        """
        If the dtype is categorical, there are two options.

        - There are only values in the data buffer.
        - There is a separate non-categorical Column encoding categorical values.

        Raises TypeError if the dtype is not categorical

        Returns the dictionary with description on how to interpret the data buffer:
            - "is_ordered" : bool, whether the ordering of dictionary indices is
                             semantically meaningful.
            - "is_dictionary" : bool, whether a mapping of
                                categorical values to other objects exists
            - "categories" : Column representing the (implicit) mapping of indices to
                             category values (e.g. an array of cat1, cat2, ...).
                             None if not a dictionary-style categorical.

        TBD: are there any other in-memory representations that are needed?
        """


class DataFrame(Protocol):
    """
    A data frame class, with only the methods required by the interchange protocol defined.

    A "data frame" represents an ordered collection of named columns.
    A column's "name" must be a unique string.
    Columns may be accessed by name or by position.

    This could be a public data frame class, or an object with the methods and
    attributes defined on this DataFrame class could be returned from the
    ``__dataframe__`` method of a public data frame class in a library adhering
    to the dataframe interchange protocol specification.
    """

    def __dataframe__(
        self, nan_as_null: bool = False, allow_copy: bool = True
    ) -> DataFrame:
        """
        Construct a new exchange object, potentially changing the parameters.

        ``nan_as_null`` is a keyword intended for the consumer to tell the
        producer to overwrite null values in the data with ``NaN``.
        It is intended for cases where the consumer does not support the bit
        mask or byte mask that is the producer's native representation.
        ``allow_copy`` is a keyword that defines whether or not the library is
        allowed to make a copy of the data. For example, copying data would be
        necessary if a library supports strided buffers, given that this protocol
        specifies contiguous buffers.
        """

    def column_names(self) -> Iterable[str]:
        """Return an iterator yielding the column names."""

    def get_column_by_name(self, name: str) -> Column:
        """Return the column whose name is the indicated name."""

    def get_chunks(self, n_chunks: int | None = None) -> Iterable[DataFrame]:
        """
        Return an iterator yielding the chunks.

        By default (None), yields the chunks that the data is stored as by the
        producer. If given, ``n_chunks`` must be a multiple of
        ``self.num_chunks()``, meaning the producer must subdivide each chunk
        before yielding it.

        Note that the producer must ensure that all columns are chunked the
        same way.
        """



from typing import Final

from altair.utils._vegafusion_data import vegafusion_data_transformer
from altair.vegalite.data import (
    DataTransformerRegistry,
    MaxRowsError,
    default_data_transformer,
    limit_rows,
    sample,
    to_csv,
    to_json,
    to_values,
)

# ==============================================================================
# VegaLite 5 data transformers
# ==============================================================================


ENTRY_POINT_GROUP: Final = "altair.vegalite.v5.data_transformer"


data_transformers = DataTransformerRegistry(entry_point_group=ENTRY_POINT_GROUP)
data_transformers.register("default", default_data_transformer)
data_transformers.register("json", to_json)
# FIXME: `to_csv` cannot accept all `DataType` https://github.com/vega/altair/issues/3441
data_transformers.register("csv", to_csv)  # type: ignore[arg-type]
data_transformers.register("vegafusion", vegafusion_data_transformer)
data_transformers.enable("default")


__all__ = (
    "MaxRowsError",
    "default_data_transformer",
    "limit_rows",
    "sample",
    "to_csv",
    "to_json",
    "to_values",
    "vegafusion_data_transformer",
)



# ruff: noqa: F401, F403
from altair.expr.core import datum

from .api import *
from .compiler import vegalite_compilers
from .data import (
    MaxRowsError,
    data_transformers,
    default_data_transformer,
    limit_rows,
    sample,
    to_csv,
    to_json,
    to_values,
)
from .display import (
    VEGA_VERSION,
    VEGAEMBED_VERSION,
    VEGALITE_VERSION,
    VegaLite,
    renderers,
)
from .schema import *
from .theme import themes



from typing import Any, Callable, Dict

from altair.utils import PluginRegistry

# ==============================================================================
# Vega-Lite to Vega compiler registry
# ==============================================================================
VegaLiteCompilerType = Callable[[Dict[str, Any]], Dict[str, Any]]


class VegaLiteCompilerRegistry(PluginRegistry[VegaLiteCompilerType, Dict[str, Any]]):
    pass



from __future__ import annotations

import struct
from typing import Any, Literal, cast, overload
from typing_extensions import TypeAlias

from ._importers import import_vl_convert, vl_version_for_vl_convert
from .html import spec_to_html

MimeBundleFormat: TypeAlias = Literal[
    "html", "json", "png", "svg", "pdf", "vega", "vega-lite"
]


@overload
def spec_to_mimebundle(
    spec: dict[str, Any],
    format: Literal["json", "vega-lite"],
    mode: Literal["vega-lite"] | None = ...,
    vega_version: str | None = ...,
    vegaembed_version: str | None = ...,
    vegalite_version: str | None = ...,
    embed_options: dict[str, Any] | None = ...,
    engine: Literal["vl-convert"] | None = ...,
    **kwargs,
) -> dict[str, dict[str, Any]]: ...
@overload
def spec_to_mimebundle(
    spec: dict[str, Any],
    format: Literal["html"],
    mode: Literal["vega-lite"] | None = ...,
    vega_version: str | None = ...,
    vegaembed_version: str | None = ...,
    vegalite_version: str | None = ...,
    embed_options: dict[str, Any] | None = ...,
    engine: Literal["vl-convert"] | None = ...,
    **kwargs,
) -> dict[str, str]: ...
@overload
def spec_to_mimebundle(
    spec: dict[str, Any],
    format: Literal["pdf", "svg", "vega"],
    mode: Literal["vega-lite"] | None = ...,
    vega_version: str | None = ...,
    vegaembed_version: str | None = ...,
    vegalite_version: str | None = ...,
    embed_options: dict[str, Any] | None = ...,
    engine: Literal["vl-convert"] | None = ...,
    **kwargs,
) -> dict[str, Any]: ...
@overload
def spec_to_mimebundle(
    spec: dict[str, Any],
    format: Literal["png"],
    mode: Literal["vega-lite"] | None = ...,
    vega_version: str | None = ...,
    vegaembed_version: str | None = ...,
    vegalite_version: str | None = ...,
    embed_options: dict[str, Any] | None = ...,
    engine: Literal["vl-convert"] | None = ...,
    **kwargs,
) -> tuple[dict[str, Any], dict[str, Any]]: ...
def spec_to_mimebundle(
    spec: dict[str, Any],
    format: MimeBundleFormat,
    mode: Literal["vega-lite"] | None = None,
    vega_version: str | None = None,
    vegaembed_version: str | None = None,
    vegalite_version: str | None = None,
    embed_options: dict[str, Any] | None = None,
    engine: Literal["vl-convert"] | None = None,
    **kwargs,
) -> dict[str, Any] | tuple[dict[str, Any], dict[str, Any]]:
    """
    Convert a vega-lite specification to a mimebundle.

    The mimebundle type is controlled by the ``format`` argument, which can be
    one of the following ['html', 'json', 'png', 'svg', 'pdf', 'vega', 'vega-lite']

    Parameters
    ----------
    spec : dict
        a dictionary representing a vega-lite plot spec
    format : string {'html', 'json', 'png', 'svg', 'pdf', 'vega', 'vega-lite'}
        the file format to be saved.
    mode : string {'vega-lite'}
        The rendering mode.
    vega_version : string
        The version of vega.js to use
    vegaembed_version : string
        The version of vegaembed.js to use
    vegalite_version : string
        The version of vegalite.js to use. Only required if mode=='vega-lite'
    embed_options : dict (optional)
        The vegaEmbed options dictionary. Defaults to the embed options set with
        alt.renderers.set_embed_options().
        (See https://github.com/vega/vega-embed for details)
    engine: string {'vl-convert'}
        the conversion engine to use for 'png', 'svg', 'pdf', and 'vega' formats
    **kwargs :
        Additional arguments will be passed to the generating function

    Returns
    -------
    output : dict
        a mime-bundle representing the image

    Note
    ----
    The png, svg, pdf, and vega outputs require the vl-convert package
    """
    # Local import to avoid circular ImportError
    from altair import renderers
    from altair.utils.display import compile_with_vegafusion, using_vegafusion

    if mode != "vega-lite":
        msg = "mode must be 'vega-lite'"
        raise ValueError(msg)

    internal_mode: Literal["vega-lite", "vega"] = mode
    if using_vegafusion():
        spec = compile_with_vegafusion(spec)
        internal_mode = "vega"

    # Default to the embed options set by alt.renderers.set_embed_options
    if embed_options is None:
        final_embed_options = renderers.options.get("embed_options", {})
    else:
        final_embed_options = embed_options

    embed_options = preprocess_embed_options(final_embed_options)

    if format in {"png", "svg", "pdf", "vega"}:
        return _spec_to_mimebundle_with_engine(
            spec,
            cast(Literal["png", "svg", "pdf", "vega"], format),
            internal_mode,
            engine=engine,
            format_locale=embed_options.get("formatLocale", None),
            time_format_locale=embed_options.get("timeFormatLocale", None),
            **kwargs,
        )
    elif format == "html":
        html = spec_to_html(
            spec,
            mode=internal_mode,
            vega_version=vega_version,
            vegaembed_version=vegaembed_version,
            vegalite_version=vegalite_version,
            embed_options=embed_options,
            **kwargs,
        )
        return {"text/html": html}
    elif format == "vega-lite":
        if vegalite_version is None:
            msg = "Must specify vegalite_version"
            raise ValueError(msg)
        return {f"application/vnd.vegalite.v{vegalite_version[0]}+json": spec}
    elif format == "json":
        return {"application/json": spec}
    else:
        msg = (
            "format must be one of "
            "['html', 'json', 'png', 'svg', 'pdf', 'vega', 'vega-lite']"
        )
        raise ValueError(msg)


def _spec_to_mimebundle_with_engine(
    spec: dict,
    format: Literal["png", "svg", "pdf", "vega"],
    mode: Literal["vega-lite", "vega"],
    format_locale: str | dict | None = None,
    time_format_locale: str | dict | None = None,
    **kwargs,
) -> Any:
    """
    Helper for Vega-Lite to mimebundle conversions that require an engine.

    Parameters
    ----------
    spec : dict
        a dictionary representing a vega-lite plot spec
    format : string {'png', 'svg', 'pdf', 'vega'}
        the format of the mimebundle to be returned
    mode : string {'vega-lite', 'vega'}
        The rendering mode.
    engine: string {'vl-convert'}
        the conversion engine to use
    format_locale : str or dict
        d3-format locale name or dictionary. Defaults to "en-US" for United States English.
        See https://github.com/d3/d3-format/tree/main/locale for available names and example
        definitions.
    time_format_locale : str or dict
        d3-time-format locale name or dictionary. Defaults to "en-US" for United States English.
        See https://github.com/d3/d3-time-format/tree/main/locale for available names and example
        definitions.
    **kwargs :
        Additional arguments will be passed to the conversion function
    """
    # Normalize the engine string (if any) by lower casing
    # and removing underscores and hyphens
    engine = kwargs.pop("engine", None)
    normalized_engine = _validate_normalize_engine(engine, format)

    if normalized_engine == "vlconvert":
        vlc = import_vl_convert()
        vl_version = vl_version_for_vl_convert()
        if format == "vega":
            if mode == "vega":
                vg = spec
            else:
                vg = vlc.vegalite_to_vega(spec, vl_version=vl_version)
            return {"application/vnd.vega.v5+json": vg}
        elif format == "svg":
            if mode == "vega":
                svg = vlc.vega_to_svg(
                    spec,
                    format_locale=format_locale,
                    time_format_locale=time_format_locale,
                )
            else:
                svg = vlc.vegalite_to_svg(
                    spec,
                    vl_version=vl_version,
                    format_locale=format_locale,
                    time_format_locale=time_format_locale,
                )
            return {"image/svg+xml": svg}
        elif format == "png":
            scale = kwargs.get("scale_factor", 1)
            # The default ppi for a PNG file is 72
            default_ppi = 72
            ppi = kwargs.get("ppi", default_ppi)
            if mode == "vega":
                png = vlc.vega_to_png(
                    spec,
                    scale=scale,
                    ppi=ppi,
                    format_locale=format_locale,
                    time_format_locale=time_format_locale,
                )
            else:
                png = vlc.vegalite_to_png(
                    spec,
                    vl_version=vl_version,
                    scale=scale,
                    ppi=ppi,
                    format_locale=format_locale,
                    time_format_locale=time_format_locale,
                )
            factor = ppi / default_ppi
            w, h = _pngxy(png)
            return {"image/png": png}, {
                "image/png": {"width": w / factor, "height": h / factor}
            }
        elif format == "pdf":
            scale = kwargs.get("scale_factor", 1)
            if mode == "vega":
                pdf = vlc.vega_to_pdf(
                    spec,
                    scale=scale,
                    format_locale=format_locale,
                    time_format_locale=time_format_locale,
                )
            else:
                pdf = vlc.vegalite_to_pdf(
                    spec,
                    vl_version=vl_version,
                    scale=scale,
                    format_locale=format_locale,
                    time_format_locale=time_format_locale,
                )
            return {"application/pdf": pdf}
        else:
            # This should be validated above
            # but raise exception for the sake of future development
            msg = f"Unexpected format {format!r}"
            raise ValueError(msg)
    else:
        # This should be validated above
        # but raise exception for the sake of future development
        msg = f"Unexpected normalized_engine {normalized_engine!r}"
        raise ValueError(msg)


def _validate_normalize_engine(
    engine: Literal["vl-convert"] | None,
    format: Literal["png", "svg", "pdf", "vega"],
) -> str:
    """
    Helper to validate and normalize the user-provided engine.

    engine : {None, 'vl-convert'}
        the user-provided engine string
    format : string {'png', 'svg', 'pdf', 'vega'}
        the format of the mimebundle to be returned
    """
    # Try to import vl_convert
    try:
        vlc = import_vl_convert()
    except ImportError:
        vlc = None

    # Normalize engine string by lower casing and removing underscores and hyphens
    normalized_engine = (
        None if engine is None else engine.lower().replace("-", "").replace("_", "")
    )

    # Validate or infer default value of normalized_engine
    if normalized_engine == "vlconvert":
        if vlc is None:
            msg = "The 'vl-convert' conversion engine requires the vl-convert-python package"
            raise ValueError(msg)
    elif normalized_engine is None:
        if vlc is not None:
            normalized_engine = "vlconvert"
        else:
            msg = (
                f"Saving charts in {format!r} format requires the vl-convert-python package: "
                "see https://altair-viz.github.io/user_guide/saving_charts.html#png-svg-and-pdf-format"
            )
            raise ValueError(msg)
    else:
        msg = f"Invalid conversion engine {engine!r}. Expected vl-convert"
        raise ValueError(msg)
    return normalized_engine


def _pngxy(data):
    """
    read the (width, height) from a PNG header.

    Taken from IPython.display
    """
    ihdr = data.index(b"IHDR")
    # next 8 bytes are width/height
    return struct.unpack(">ii", data[ihdr + 4 : ihdr + 12])


def preprocess_embed_options(embed_options: dict) -> dict:
    """
    Preprocess embed options to a form compatible with Vega Embed.

    Parameters
    ----------
    embed_options : dict
        The embed options dictionary to preprocess.

    Returns
    -------
    embed_opts : dict
        The preprocessed embed options dictionary.
    """
    embed_options = (embed_options or {}).copy()

    # Convert locale strings to objects compatible with Vega Embed using vl-convert
    format_locale = embed_options.get("formatLocale", None)
    if isinstance(format_locale, str):
        vlc = import_vl_convert()
        embed_options["formatLocale"] = vlc.get_format_locale(format_locale)

    time_format_locale = embed_options.get("timeFormatLocale", None)
    if isinstance(time_format_locale, str):
        vlc = import_vl_convert()
        embed_options["timeFormatLocale"] = vlc.get_time_format_locale(
            time_format_locale
        )

    return embed_options



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Final

from altair.utils.mimebundle import spec_to_mimebundle
from altair.vegalite.display import (
    Displayable,
    HTMLRenderer,
    RendererRegistry,
    default_renderer_base,
    json_renderer_base,
)

from .schema import SCHEMA_VERSION

if TYPE_CHECKING:
    from altair.vegalite.display import DefaultRendererReturnType


VEGALITE_VERSION: Final = SCHEMA_VERSION.lstrip("v")
VEGA_VERSION: Final = "5"
VEGAEMBED_VERSION: Final = "6"


# ==============================================================================
# VegaLite v5 renderer logic
# ==============================================================================


# The MIME type for Vega-Lite 5.x releases.
VEGALITE_MIME_TYPE: Final = "application/vnd.vegalite.v5+json"

# The MIME type for Vega 5.x releases.
VEGA_MIME_TYPE: Final = "application/vnd.vega.v5+json"

# The entry point group that can be used by other packages to declare other
# renderers that will be auto-detected. Explicit registration is also
# allowed by the PluginRegistery API.
ENTRY_POINT_GROUP: Final = "altair.vegalite.v5.renderer"

# The display message when rendering fails
DEFAULT_DISPLAY: Final = f"""\
<VegaLite {VEGALITE_VERSION.split('.')[0]} object>

If you see this message, it means the renderer has not been properly enabled
for the frontend that you are using. For more information, see
https://altair-viz.github.io/user_guide/display_frontends.html#troubleshooting
"""

renderers = RendererRegistry(entry_point_group=ENTRY_POINT_GROUP)

here = str(Path(__file__).parent)


def mimetype_renderer(spec: dict, **metadata) -> DefaultRendererReturnType:
    return default_renderer_base(spec, VEGALITE_MIME_TYPE, DEFAULT_DISPLAY, **metadata)


def json_renderer(spec: dict, **metadata) -> DefaultRendererReturnType:
    return json_renderer_base(spec, DEFAULT_DISPLAY, **metadata)


def png_renderer(spec: dict, **metadata) -> dict[str, bytes]:
    # To get proper return value type, would need to write complex
    # overload signatures for spec_to_mimebundle based on `format`
    return spec_to_mimebundle(  # type: ignore[return-value]
        spec,
        format="png",
        mode="vega-lite",
        vega_version=VEGA_VERSION,
        vegaembed_version=VEGAEMBED_VERSION,
        vegalite_version=VEGALITE_VERSION,
        **metadata,
    )


def svg_renderer(spec: dict, **metadata) -> dict[str, str]:
    # To get proper return value type, would need to write complex
    # overload signatures for spec_to_mimebundle based on `format`
    return spec_to_mimebundle(
        spec,
        format="svg",
        mode="vega-lite",
        vega_version=VEGA_VERSION,
        vegaembed_version=VEGAEMBED_VERSION,
        vegalite_version=VEGALITE_VERSION,
        **metadata,
    )


def jupyter_renderer(spec: dict, **metadata):
    """Render chart using the JupyterChart Jupyter Widget."""
    from altair import Chart, JupyterChart

    # Configure offline mode
    offline = metadata.get("offline", False)

    # mypy doesn't see the enable_offline class method for some reason
    JupyterChart.enable_offline(offline=offline)  # type: ignore[attr-defined]

    # propagate embed options
    embed_options = metadata.get("embed_options", None)

    # Need to ignore attr-defined mypy rule because mypy doesn't see _repr_mimebundle_
    # conditionally defined in AnyWidget
    return JupyterChart(
        chart=Chart.from_dict(spec), embed_options=embed_options
    )._repr_mimebundle_()  # type: ignore[attr-defined]


def browser_renderer(
    spec: dict, offline=False, using=None, port=0, **metadata
) -> dict[str, str]:
    from altair.utils._show import open_html_in_browser

    if offline:
        metadata["template"] = "inline"
    mimebundle = spec_to_mimebundle(
        spec,
        format="html",
        mode="vega-lite",
        vega_version=VEGA_VERSION,
        vegaembed_version=VEGAEMBED_VERSION,
        vegalite_version=VEGALITE_VERSION,
        **metadata,
    )

    if isinstance(mimebundle, tuple):
        mimebundle = mimebundle[0]

    html = mimebundle["text/html"]
    open_html_in_browser(html, using=using, port=port)
    return {}


html_renderer = HTMLRenderer(
    mode="vega-lite",
    template="universal",
    vega_version=VEGA_VERSION,
    vegaembed_version=VEGAEMBED_VERSION,
    vegalite_version=VEGALITE_VERSION,
)

renderers.register("default", html_renderer)
renderers.register("html", html_renderer)
renderers.register("colab", html_renderer)
renderers.register("kaggle", html_renderer)
renderers.register("zeppelin", html_renderer)
renderers.register("mimetype", mimetype_renderer)
renderers.register("jupyterlab", mimetype_renderer)
renderers.register("nteract", mimetype_renderer)
renderers.register("json", json_renderer)
renderers.register("png", png_renderer)
renderers.register("svg", svg_renderer)
renderers.register("jupyter", jupyter_renderer)
renderers.register("browser", browser_renderer)
renderers.enable("default")


class VegaLite(Displayable):
    """An IPython/Jupyter display class for rendering VegaLite 5."""

    renderers = renderers
    schema_path = (__name__, "schema/vega-lite-schema.json")


def vegalite(spec: dict, validate: bool = True) -> None:
    """
    Render and optionally validate a VegaLite 5 spec.

    This will use the currently enabled renderer to render the spec.

    Parameters
    ----------
    spec: dict
        A fully compliant VegaLite 5 spec, with the data portion fully processed.
    validate: bool
        Should the spec be validated against the VegaLite 5 schema?
    """
    from IPython.display import display

    display(VegaLite(spec, validate=validate))



"""Tools for enabling and registering chart themes."""

from __future__ import annotations

from typing import TYPE_CHECKING, Final, Literal, get_args

from altair.utils.theme import ThemeRegistry
from altair.vegalite.v5.schema._typing import VegaThemes

if TYPE_CHECKING:
    import sys

    if sys.version_info >= (3, 10):
        from typing import TypeAlias
    else:
        from typing_extensions import TypeAlias

AltairThemes: TypeAlias = Literal["default", "opaque"]
VEGA_THEMES: list[str] = list(get_args(VegaThemes))


class VegaTheme:
    """Implementation of a builtin vega theme."""

    def __init__(self, theme: str) -> None:
        self.theme = theme

    def __call__(self) -> dict[str, dict[str, dict[str, str | int]]]:
        return {
            "usermeta": {"embedOptions": {"theme": self.theme}},
            "config": {"view": {"continuousWidth": 300, "continuousHeight": 300}},
        }

    def __repr__(self) -> str:
        return f"VegaTheme({self.theme!r})"


# The entry point group that can be used by other packages to declare other
# themes that will be auto-detected. Explicit registration is also
# allowed by the PluginRegistry API.
ENTRY_POINT_GROUP: Final = "altair.vegalite.v5.theme"
themes = ThemeRegistry(entry_point_group=ENTRY_POINT_GROUP)

themes.register(
    "default",
    lambda: {"config": {"view": {"continuousWidth": 300, "continuousHeight": 300}}},
)
themes.register(
    "opaque",
    lambda: {
        "config": {
            "background": "white",
            "view": {"continuousWidth": 300, "continuousHeight": 300},
        }
    },
)
themes.register("none", dict)

for theme in VEGA_THEMES:
    themes.register(theme, VegaTheme(theme))

themes.enable("default")



"""Utilities for registering and working with themes."""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Callable

from .plugin_registry import PluginRegistry

if sys.version_info >= (3, 11):
    from typing import LiteralString
else:
    from typing_extensions import LiteralString

if TYPE_CHECKING:
    from altair.utils.plugin_registry import PluginEnabler
    from altair.vegalite.v5.theme import AltairThemes, VegaThemes

ThemeType = Callable[..., dict]


class ThemeRegistry(PluginRegistry[ThemeType, dict]):
    def enable(
        self, name: LiteralString | AltairThemes | VegaThemes | None = None, **options
    ) -> PluginEnabler:
        """
        Enable a theme by name.

        This can be either called directly, or used as a context manager.

        Parameters
        ----------
        name : string (optional)
            The name of the theme to enable. If not specified, then use the
            current active name.
        **options :
            Any additional parameters will be passed to the theme as keyword
            arguments

        Returns
        -------
        PluginEnabler:
            An object that allows enable() to be used as a context manager

        Notes
        -----
        Default `vega` themes can be previewed at https://vega.github.io/vega-themes/
        """
        return super().enable(name, **options)
